from typing_extensions import Literal

from ome_types._base_type import OMEType

from .shape import Shape


class Polygon(Shape, OMEType):
    """The Polygon defines closed shapes formed of straight lines.

    Note: Polygon uses counterclockwise winding (this is the default OpenGL
    behavior)

    Parameters
    ----------
    id : ShapeID
    points : str
        The points of the Polygon are defined as a list of comma separated x,y
        coordinates seperated by spaces like "x1,y1 x2,y2 x3,y3" e.g. "0,0 1,2
        3,5"
    annotation_ref : AnnotationRef, optional
    fill_color : Color, optional
        The color of the fill - encoded as RGBA The value "-1" is #FFFFFFFF so
        solid white (it is a signed 32 bit value) NOTE: Prior to the 2012-06
        schema the default value was incorrect and produced a transparent red
        not solid white.
    fill_rule : FillRule, optional
    font_family : FontFamily, optional
    font_size : NonNegativeInt, optional
        Size of the font. Units are set by FontSizeUnit.
    font_size_unit : UnitsLength, optional
        The units used for the font size.
    font_style : FontStyle, optional
    locked : bool, optional
        Controls whether the shape is locked and read only, true is locked,
        false is editable.
    stroke_color : Color, optional
        The color of the stroke  - encoded as RGBA The value "-1" is #FFFFFFFF
        so solid white (it is a signed 32 bit value) NOTE: Prior to the
        2012-06 schema the default value was incorrect and produced a
        transparent red not solid white.
    stroke_dash_array : str, optional
        e.g. "none", "10 20 30 10"
    stroke_width : float, optional
        The width of the stroke. Units are set by StrokeWidthUnit.
    stroke_width_unit : UnitsLength, optional
        The units used for the stroke width.
    text : str, optional
    the_c : NonNegativeInt, optional
        The channel the ROI applies to. If not specified then the ROI applies
        to all the channels of the image. This is numbered from 0.
    the_t : NonNegativeInt, optional
        The timepoint the ROI applies to. If not specified then the ROI
        applies to all the timepoints of the image. This is numbered from 0.
    the_z : NonNegativeInt, optional
        The z-section the ROI applies to. If not specified then the ROI
        applies to all the z-sections of the image. This is numbered from 0.
    transform : AffineTransform, optional
        This is a matrix used to transform the shape. The element has 6
        xsd:float attributes. If the element is present then all 6 values must
        be included.
    """

    kind: Literal["polygon"] = "polygon"
    points: str



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import FolderID


class FolderRef(Reference, OMEType):
    """The FolderRef element refers to a Folder by specifying the Folder ID
    attribute.

    One or more FolderRef elements may be listed within the Folder element to
    specify what Folders the Folder contains. This tree hierarchy must be acyclic.

    Parameters
    ----------
    id : FolderID
    """

    id: FolderID



import warnings

from ._xmlschema import *  # noqa

warnings.warn(
    "Direct import from ome_types.schema is deprecated. "
    "Please import convenience functions directly from ome_types. "
    "This will raise an error in the future.",
    FutureWarning,
    stacklevel=2,
)



import re
from enum import Enum

from pydantic import color
from pydantic.types import ConstrainedFloat, ConstrainedInt, ConstrainedStr


class base64Binary(ConstrainedStr):
    pass


class Binning(Enum):
    """Represents the number of pixels that are combined to form larger pixels.

    {used:CCD,EMCCD}.
    """

    EIGHTBYEIGHT = "8x8"
    FOURBYFOUR = "4x4"
    ONEBYONE = "1x1"
    OTHER = "Other"
    TWOBYTWO = "2x2"


class Color(color.Color):
    def __init__(self, val: color.ColorType) -> None:
        if isinstance(val, int):
            val = self._int2tuple(val)
        super().__init__(val)

    @classmethod
    def _int2tuple(cls, val: int):
        return (val >> 24 & 255, val >> 16 & 255, val >> 8 & 255, (val & 255) / 255)

    def as_int32(self) -> int:
        r, g, b, *a = self.as_rgb_tuple()
        v = r << 24 | g << 16 | b << 8 | int((a[0] if a else 1) * 255)
        if v < 2**32 // 2:
            return v
        return v - 2**32

    def __eq__(self, o: object) -> bool:
        if isinstance(o, Color):
            return self.as_int32() == o.as_int32()
        return False

    def __int__(self) -> int:
        return self.as_int32()


class FontFamily(Enum):
    """The font family used to draw the text.

    [enumeration] Note: these values are all lower case so they match the standard
    HTML/CSS values. "fantasy" has been included for completeness we do not
    recommended its regular use.
    """

    CURSIVE = "cursive"
    FANTASY = "fantasy"
    MONOSPACE = "monospace"
    SANSSERIF = "sans-serif"
    SERIF = "serif"


class Hex40(ConstrainedStr):
    min_length = 40
    max_length = 40


class LSID(ConstrainedStr):
    regex = re.compile(r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:\S+:\S+)|(\S+:\S+)")


class Marker(Enum):
    """Shape of marker on the end of a line.

    [enumeration].
    """

    ARROW = "Arrow"


class NamingConvention(Enum):
    """Predefined list of values for the well labels."""

    LETTER = "letter"
    NUMBER = "number"


class NonNegativeFloat(ConstrainedFloat):
    ge = 0.0


class NonNegativeInt(ConstrainedInt):
    ge = 0


class NonNegativeLong(ConstrainedInt):
    ge = 0


class PercentFraction(ConstrainedFloat):
    le = 1.0
    ge = 0.0


class PixelType(Enum):
    """The number size/kind used to represent a pixel."""

    BIT = "bit"
    COMPLEXDOUBLE = "double-complex"
    COMPLEXFLOAT = "complex"
    DOUBLE = "double"
    FLOAT = "float"
    INT16 = "int16"
    INT32 = "int32"
    INT8 = "int8"
    UINT16 = "uint16"
    UINT32 = "uint32"
    UINT8 = "uint8"


class PositiveFloat(ConstrainedFloat):
    gt = 0.0


class PositiveInt(ConstrainedInt):
    ge = 1


class UnitsAngle(Enum):
    """The units used to represent an angle."""

    DEGREE = "deg"
    GRADIAN = "gon"
    RADIAN = "rad"


class UnitsElectricPotential(Enum):
    """The units used to represent an electric potential."""

    ATTOVOLT = "aV"
    CENTIVOLT = "cV"
    DECAVOLT = "daV"
    DECIVOLT = "dV"
    EXAVOLT = "EV"
    FEMTOVOLT = "fV"
    GIGAVOLT = "GV"
    HECTOVOLT = "hV"
    KILOVOLT = "kV"
    MEGAVOLT = "MV"
    MICROVOLT = "µV"
    MILLIVOLT = "mV"
    NANOVOLT = "nV"
    PETAVOLT = "PV"
    PICOVOLT = "pV"
    TERAVOLT = "TV"
    VOLT = "V"
    YOCTOVOLT = "yV"
    YOTTAVOLT = "YV"
    ZEPTOVOLT = "zV"
    ZETTAVOLT = "ZV"


class UnitsFrequency(Enum):
    """The units used to represent frequency."""

    ATTOHERTZ = "aHz"
    CENTIHERTZ = "cHz"
    DECAHERTZ = "daHz"
    DECIHERTZ = "dHz"
    EXAHERTZ = "EHz"
    FEMTOHERTZ = "fHz"
    GIGAHERTZ = "GHz"
    HECTOHERTZ = "hHz"
    HERTZ = "Hz"
    KILOHERTZ = "kHz"
    MEGAHERTZ = "MHz"
    MICROHERTZ = "µHz"
    MILLIHERTZ = "mHz"
    NANOHERTZ = "nHz"
    PETAHERTZ = "PHz"
    PICOHERTZ = "pHz"
    TERAHERTZ = "THz"
    YOCTOHERTZ = "yHz"
    YOTTAHERTZ = "YHz"
    ZEPTOHERTZ = "zHz"
    ZETTAHERTZ = "ZHz"


class UnitsLength(Enum):
    """The units used to represent a length."""

    ANGSTROM = "Å"
    ASTRONOMICALUNIT = "ua"
    ATTOMETER = "am"
    CENTIMETER = "cm"
    DECAMETER = "dam"
    DECIMETER = "dm"
    EXAMETER = "Em"
    FEMTOMETER = "fm"
    FOOT = "ft"
    GIGAMETER = "Gm"
    HECTOMETER = "hm"
    INCH = "in"
    KILOMETER = "km"
    LIGHTYEAR = "ly"
    LINE = "li"
    MEGAMETER = "Mm"
    METER = "m"
    MICROMETER = "µm"
    MILE = "mi"
    MILLIMETER = "mm"
    NANOMETER = "nm"
    PARSEC = "pc"
    PETAMETER = "Pm"
    PICOMETER = "pm"
    PIXEL = "pixel"
    POINT = "pt"
    REFERENCEFRAME = "reference frame"
    TERAMETER = "Tm"
    THOU = "thou"
    YARD = "yd"
    YOCTOMETER = "ym"
    YOTTAMETER = "Ym"
    ZEPTOMETER = "zm"
    ZETTAMETER = "Zm"


class UnitsPower(Enum):
    """The units used to represent power."""

    ATTOWATT = "aW"
    CENTIWATT = "cW"
    DECAWATT = "daW"
    DECIWATT = "dW"
    EXAWATT = "EW"
    FEMTOWATT = "fW"
    GIGAWATT = "GW"
    HECTOWATT = "hW"
    KILOWATT = "kW"
    MEGAWATT = "MW"
    MICROWATT = "µW"
    MILLIWATT = "mW"
    NANOWATT = "nW"
    PETAWATT = "PW"
    PICOWATT = "pW"
    TERAWATT = "TW"
    WATT = "W"
    YOCTOWATT = "yW"
    YOTTAWATT = "YW"
    ZEPTOWATT = "zW"
    ZETTAWATT = "ZW"


class UnitsPressure(Enum):
    """The units used to represent a pressure."""

    ATMOSPHERE = "atm"
    ATTOPASCAL = "aPa"
    BAR = "bar"
    CENTIBAR = "cbar"
    CENTIPASCAL = "cPa"
    DECAPASCAL = "daPa"
    DECIBAR = "dbar"
    DECIPASCAL = "dPa"
    EXAPASCAL = "EPa"
    FEMTOPASCAL = "fPa"
    GIGAPASCAL = "GPa"
    HECTOPASCAL = "hPa"
    KILOBAR = "kbar"
    KILOPASCAL = "kPa"
    MEGABAR = "Mbar"
    MEGAPASCAL = "MPa"
    MICROPASCAL = "µPa"
    MILLIBAR = "mbar"
    MILLIPASCAL = "mPa"
    MILLITORR = "mTorr"
    MMHG = "mm Hg"
    NANOPASCAL = "nPa"
    PASCAL = "Pa"
    PETAPASCAL = "PPa"
    PICOPASCAL = "pPa"
    PSI = "psi"
    TERAPASCAL = "TPa"
    TORR = "Torr"
    YOCTOPASCAL = "yPa"
    YOTTAPASCAL = "YPa"
    ZEPTOPASCAL = "zPa"
    ZETTAPASCAL = "ZPa"


class UnitsTemperature(Enum):
    """The units used to represent a temperature."""

    CELSIUS = "°C"
    FAHRENHEIT = "°F"
    KELVIN = "K"
    RANKINE = "°R"


class UnitsTime(Enum):
    """The units used to represent a time interval."""

    ATTOSECOND = "as"
    CENTISECOND = "cs"
    DAY = "d"
    DECASECOND = "das"
    DECISECOND = "ds"
    EXASECOND = "Es"
    FEMTOSECOND = "fs"
    GIGASECOND = "Gs"
    HECTOSECOND = "hs"
    HOUR = "h"
    KILOSECOND = "ks"
    MEGASECOND = "Ms"
    MICROSECOND = "µs"
    MILLISECOND = "ms"
    MINUTE = "min"
    NANOSECOND = "ns"
    PETASECOND = "Ps"
    PICOSECOND = "ps"
    SECOND = "s"
    TERASECOND = "Ts"
    YOCTOSECOND = "ys"
    YOTTASECOND = "Ys"
    ZEPTOSECOND = "zs"
    ZETTASECOND = "Zs"


class UniversallyUniqueIdentifier(ConstrainedStr):
    regex = re.compile(
        r"(urn:uuid:[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12})"
    )


class AnnotationID(LSID):
    regex = re.compile(
        r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Annotation:\S+)|(Annotation:\S+)"
    )

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v) -> "AnnotationID":
        if not cls.regex.match(v):
            search = cls.regex.search(v)
            if search:
                import warnings

                new_v = search.group()
                warnings.warn(f"Casting invalid AnnotationID {v!r} to {new_v!r}")
                v = new_v
        return v


class ChannelID(LSID):
    regex = re.compile(r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Channel:\S+)|(Channel:\S+)")


class DatasetID(LSID):
    regex = re.compile(r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Dataset:\S+)|(Dataset:\S+)")


class DetectorID(LSID):
    regex = re.compile(
        r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Detector:\S+)|(Detector:\S+)"
    )


class DichroicID(LSID):
    regex = re.compile(
        r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Dichroic:\S+)|(Dichroic:\S+)"
    )


class ExperimenterGroupID(LSID):
    regex = re.compile(
        r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:ExperimenterGroup:\S+)|(ExperimenterGroup:\S+)"
    )


class ExperimenterID(LSID):
    regex = re.compile(
        r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Experimenter:\S+)|(Experimenter:\S+)"
    )


class ExperimentID(LSID):
    regex = re.compile(
        r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Experiment:\S+)|(Experiment:\S+)"
    )


class FilterID(LSID):
    regex = re.compile(r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Filter:\S+)|(Filter:\S+)")


class FilterSetID(LSID):
    regex = re.compile(
        r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:FilterSet:\S+)|(FilterSet:\S+)"
    )


class FolderID(LSID):
    regex = re.compile(r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Folder:\S+)|(Folder:\S+)")


class ImageID(LSID):
    regex = re.compile(r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Image:\S+)|(Image:\S+)")


class InstrumentID(LSID):
    regex = re.compile(
        r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Instrument:\S+)|(Instrument:\S+)"
    )


class LightSourceID(LSID):
    regex = re.compile(
        r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:LightSource:\S+)|(LightSource:\S+)"
    )


class MicrobeamManipulationID(LSID):
    regex = re.compile(
        r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:MicrobeamManipulation:\S+)|(MicrobeamManipulation:\S+)"
    )


class ModuleID(LSID):
    regex = re.compile(r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Module:\S+)|(Module:\S+)")


class ObjectiveID(LSID):
    regex = re.compile(
        r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Objective:\S+)|(Objective:\S+)"
    )


class PixelsID(LSID):
    regex = re.compile(r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Pixels:\S+)|(Pixels:\S+)")


class PlateAcquisitionID(LSID):
    regex = re.compile(
        r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:PlateAcquisition:\S+)|(PlateAcquisition:\S+)"
    )


class PlateID(LSID):
    regex = re.compile(r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Plate:\S+)|(Plate:\S+)")


class ProjectID(LSID):
    regex = re.compile(r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Project:\S+)|(Project:\S+)")


class ReagentID(LSID):
    regex = re.compile(r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Reagent:\S+)|(Reagent:\S+)")


class ROIID(LSID):
    regex = re.compile(r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:\S+)|(\S+)")


class ScreenID(LSID):
    regex = re.compile(r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Screen:\S+)|(Screen:\S+)")


class ShapeID(LSID):
    regex = re.compile(r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Shape:\S+)|(Shape:\S+)")


class WellID(LSID):
    regex = re.compile(r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:Well:\S+)|(Well:\S+)")


class WellSampleID(LSID):
    regex = re.compile(
        r"(urn:lsid:([\w\-\.]+\.[\w\-\.]+)+:WellSample:\S+)|(WellSample:\S+)"
    )



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import ExperimenterGroupID


class ExperimenterGroupRef(Reference, OMEType):
    """This empty element has a reference (the ExperimenterGroup ID attribute) to a
    ExperimenterGroup defined within OME.

    Parameters
    ----------
    id : ExperimenterGroupID
    """

    id: ExperimenterGroupID



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import ReagentID


class ReagentRef(Reference, OMEType):
    """ReagentRef.

    Parameters
    ----------
    id : ReagentID
    """

    id: ReagentID



from pathlib import Path

URI_OME = "http://www.openmicroscopy.org/Schemas/OME/2016-06"
NS_OME = "{" + URI_OME + "}"
NS_XSI = "{http://www.w3.org/2001/XMLSchema-instance}"
SCHEMA_LOC_OME = f"{URI_OME} {URI_OME}/ome.xsd"


OME_2016_06_XSD = str(Path(__file__).parent / "ome-2016-06.xsd")



from enum import Enum
from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .bin_data import BinData
from .channel import Channel
from .plane import Plane
from .simple_types import (
    PixelsID,
    PixelType,
    PositiveFloat,
    PositiveInt,
    UnitsLength,
    UnitsTime,
)
from .tiff_data import TiffData


class DimensionOrder(Enum):
    """The order in which the individual planes of data are interleaved."""

    XYCTZ = "XYCTZ"
    XYCZT = "XYCZT"
    XYTCZ = "XYTCZ"
    XYTZC = "XYTZC"
    XYZCT = "XYZCT"
    XYZTC = "XYZTC"


class Pixels(OMEType):
    """Pixels is going to be removed in the future, but it is still required.

    This is just notice that the contents of Pixels will be moved up to Image in a
    future release. This is because there has only been 1 Pixels object in each
    Image for some time. The concept of multiple Pixels sets for one Image failed
    to take off. It is therefore redundant.

    The Image will be unreadable if any of the required Pixel attributes are
    missing.

    The Pixels themselves can be stored within the OME-XML compressed by plane,
    and encoded in Base64. Or the Pixels may be stored in TIFF format.

    The Pixels element should contain a list of BinData or TiffData, each
    containing a single plane of pixels. These Pixels elements, when read in
    document order, must produce a 5-D pixel array of the size specified in this
    element, and in the dimension order specified by 'DimensionOrder'.

    All of the values in the Pixels object when present should match the same
    value stored in any associated TIFF format (e.g. SizeX should be the same).
    Where there is a mismatch our readers will take the value from the TIFF
    structure as overriding the value in the OME-XML. This is simply a pragmatic
    decision as it increases the likelihood of reading data from a slightly
    incorrect file.

    Parameters
    ----------
    dimension_order : DimensionOrder
        The order in which the individual planes of data are interleaved.
    id : PixelsID
    size_c : PositiveInt
        Dimensional size of pixel data array
    size_t : PositiveInt
        Dimensional size of pixel data array
    size_x : PositiveInt
        Dimensional size of pixel data array
    size_y : PositiveInt
        Dimensional size of pixel data array
    size_z : PositiveInt
        Dimensional size of pixel data array
    type : PixelType
        The variable type used to represent each pixel in the image.
    big_endian : bool, optional
        This is true if the pixels data was written in BigEndian order.  If
        this value is present it should match the value used in BinData or
        TiffData. If it does not a reader should honour the value used in the
        BinData or TiffData. This values is useful for MetadataOnly files and
        is to allow for future storage solutions.
    bin_data : BinData, optional
    channels : Channel, optional
    interleaved : bool, optional
        How the channels are arranged within the data block: true if channels
        are stored RGBRGBRGB...; false if channels are stored
        RRR...GGG...BBB...
    metadata_only : bool, optional
    physical_size_x : PositiveFloat, optional
        Physical size of a pixel. Units are set by PhysicalSizeXUnit.
    physical_size_x_unit : UnitsLength, optional
        The units of the physical size of a pixel - default:microns.
    physical_size_y : PositiveFloat, optional
        Physical size of a pixel. Units are set by PhysicalSizeYUnit.
    physical_size_y_unit : UnitsLength, optional
        The units of the physical size of a pixel - default:microns.
    physical_size_z : PositiveFloat, optional
        Physical size of a pixel. Units are set by PhysicalSizeZUnit.
    physical_size_z_unit : UnitsLength, optional
        The units of the physical size of a pixel - default:microns.
    planes : Plane, optional
    significant_bits : PositiveInt, optional
        The number of bits within the type storing each pixel that are
        significant. e.g. you can store 12 bit data within a 16 bit type. This
        does not reduce the storage requirements but can be a useful indicator
        when processing or viewing the image data.
    tiff_data_blocks : TiffData, optional
    time_increment : float, optional
        TimeIncrement is used for time series that have a global timing
        specification instead of per-timepoint timing info. For example in a
        video stream. Units are set by TimeIncrementUnit.
    time_increment_unit : UnitsTime, optional
        The units of the TimeIncrement - default:seconds.
    """

    dimension_order: DimensionOrder
    id: PixelsID
    size_c: PositiveInt
    size_t: PositiveInt
    size_x: PositiveInt
    size_y: PositiveInt
    size_z: PositiveInt
    type: PixelType
    big_endian: Optional[bool] = None
    bin_data: List[BinData] = Field(default_factory=list)
    channels: List[Channel] = Field(default_factory=list)
    interleaved: Optional[bool] = None
    metadata_only: bool = False
    physical_size_x: Optional[PositiveFloat] = None
    physical_size_x_unit: Optional[UnitsLength] = UnitsLength("µm")
    physical_size_y: Optional[PositiveFloat] = None
    physical_size_y_unit: Optional[UnitsLength] = UnitsLength("µm")
    physical_size_z: Optional[PositiveFloat] = None
    physical_size_z_unit: Optional[UnitsLength] = UnitsLength("µm")
    planes: List[Plane] = Field(default_factory=list)
    significant_bits: Optional[PositiveInt] = None
    tiff_data_blocks: List[TiffData] = Field(default_factory=list)
    time_increment: Optional[float] = None
    time_increment_unit: Optional[UnitsTime] = UnitsTime("s")



import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, Union, cast
from warnings import warn

from typing_extensions import Protocol

from .model import OME

if TYPE_CHECKING:
    from ._xmlschema import XMLSourceType


class Parser(Protocol):
    # Used for type checks on xml parsers
    def __call__(
        self, path_or_str: Union[Path, str, bytes], validate: Optional[bool] = False
    ) -> Dict[str, Any]:
        ...


def to_dict(
    xml: Union[Path, str, bytes],
    *,
    parser: Union[Parser, str, None] = None,
    validate: Optional[bool] = None,
) -> Dict[str, Any]:
    """Convert OME XML to dict.

    Parameters
    ----------
    xml : Union[Path, str, bytes]
        XML string or path to XML file.
    parser : Union[Parser, str]
        Either a parser callable with signature:
        `(path_or_str: Union[Path, str, bytes], validate: Optional[bool] = False) ->
        Dict`, or a string.  If a string, must be either 'lxml' or 'xmlschema'. by
        default "lxml"
    validate : Optional[bool], optional
        Whether to validate XML as valid OME XML, by default (`None`), the choices is
        left to the parser (which is `False` for the lxml parser)

    Returns
    -------
    Dict[str, Any]
        OME model dict.

    Raises
    ------
    KeyError
        If `parser` is a string, and not one of `'lxml'` or `'xmlschema'`
    """
    if parser is None:
        warn(
            "The default XML parser will be changing from 'xmlschema' to 'lxml' in "
            "version 0.4.0.  To silence this warning, please provide the `parser` "
            "argument, specifying either 'lxml' (to opt into the new behavior), or"
            "'xmlschema' (to retain the old behavior).",
            FutureWarning,
            stacklevel=2,
        )
        parser = "xmlschema"

    if isinstance(parser, str):
        if parser == "lxml":
            from ._lxml import xml2dict

            parser = cast(Parser, xml2dict)
        elif parser == "xmlschema":
            from ._xmlschema import xmlschema2dict

            parser = cast(Parser, xmlschema2dict)
        else:
            raise KeyError("parser string must be one of {'lxml', 'xmlschema'}")

    if validate is True:
        validate_xml(xml)

    d = parser(xml)
    for key in list(d.keys()):
        if key.startswith(("xml", "xsi")):
            d.pop(key)
    return d


def from_xml(
    xml: Union[Path, str, bytes],
    *,
    parser: Union[Parser, str, None] = None,
    validate: Optional[bool] = None,
) -> OME:
    """Generate OME metadata object from XML string or path.

    Parameters
    ----------
    xml : Union[Path, str, bytes]
        XML string or path to XML file.
    parser : Union[Parser, str]
        Either a parser callable with signature: `(path_or_str: Union[Path, str, bytes],
        validate: Optional[bool] = False) -> Dict`, or a string.  If a string, must be
        either 'lxml' or 'xmlschema'. by default "lxml"
    validate : Optional[bool], optional
        Whether to validate XML as valid OME XML, by default (`None`), the choices is
        left to the parser (which is `False` for the lxml parser)


    Returns
    -------
    ome: ome_types.model.ome.OME
        ome_types.OME metadata object
    """
    d = to_dict(os.fspath(xml), parser=parser, validate=validate)
    return OME(**d)


def from_tiff(
    path: Union[Path, str],
    *,
    parser: Union[Parser, str, None] = None,
    validate: Optional[bool] = True,
) -> OME:
    """Generate OME metadata object from OME-TIFF path.

    This will use the first ImageDescription tag found in the TIFF header.

    Parameters
    ----------
    path : Union[Path, str]
        Path to OME TIFF.
    parser : Union[Parser, str]
        Either a parser callable with signature: `(path_or_str: Union[Path, str, bytes],
        validate: Optional[bool] = False) -> Dict`, or a string.  If a string, must be
        either 'lxml' or 'xmlschema'. by default "lxml"
    validate : Optional[bool], optional
        Whether to validate XML as valid OME XML, by default (`None`), the choices is
        left to the parser (which is `False` for the lxml parser)


    Returns
    -------
    ome: ome_types.model.ome.OME
        ome_types.OME metadata object

    Raises
    ------
    ValueError
        If the TIFF file has no OME metadata.
    """
    return from_xml(_tiff2xml(path), parser=parser, validate=validate)


def _tiff2xml(path: Union[Path, str]) -> bytes:
    """Extract OME XML from OME-TIFF path.

    This will use the first ImageDescription tag found in the TIFF header.

    Parameters
    ----------
    path : Union[Path, str]
        Path to OME TIFF.

    Returns
    -------
    xml : str
        OME XML

    Raises
    ------
    ValueError
        If the TIFF file has no OME metadata.
    """
    from struct import unpack

    with Path(path).open(mode="rb") as fh:
        try:
            offsetsize, offsetformat, tagnosize, tagnoformat, tagsize, codeformat = {
                b"II*\0": (4, "<I", 2, "<H", 12, "<H"),
                b"MM\0*": (4, ">I", 2, ">H", 12, ">H"),
                b"II+\0": (8, "<Q", 8, "<Q", 20, "<H"),
                b"MM\0+": (8, ">Q", 8, ">Q", 20, ">H"),
            }[fh.read(4)]
        except KeyError as e:
            raise ValueError(f"{path!r} does not have a recognized TIFF header") from e

        fh.read(4 if offsetsize == 8 else 0)
        fh.seek(unpack(offsetformat, fh.read(offsetsize))[0])
        for _ in range(unpack(tagnoformat, fh.read(tagnosize))[0]):
            tagstruct = fh.read(tagsize)
            if unpack(codeformat, tagstruct[:2])[0] == 270:
                size = unpack(offsetformat, tagstruct[4 : 4 + offsetsize])[0]
                if size <= offsetsize:
                    desc = tagstruct[4 + offsetsize : 4 + offsetsize + size]
                    break
                fh.seek(unpack(offsetformat, tagstruct[-offsetsize:])[0])
                desc = fh.read(size)
                break
        else:
            raise ValueError(f"No OME metadata found in file: {path}")
    if desc[-1] == 0:
        desc = desc[:-1]
    return desc


def to_xml(ome: OME, **kwargs: Any) -> str:
    """
    Dump an OME object to string.

    Parameters
    ----------
    ome: OME
        OME object to dump.
    **kwargs
        Extra kwargs to pass to ElementTree.tostring.

    Returns
    -------
    ome_string: str
        The XML string of the OME object.
    """
    from ._xmlschema import to_xml

    return to_xml(ome, **kwargs)


def validate_xml(xml: "XMLSourceType", schema: Any = None) -> None:
    from ._xmlschema import validate

    validate(xml, schema=schema)



from datetime import datetime
from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .experiment_ref import ExperimentRef
from .experimenter_group_ref import ExperimenterGroupRef
from .experimenter_ref import ExperimenterRef
from .imaging_environment import ImagingEnvironment
from .instrument_ref import InstrumentRef
from .microbeam_manipulation_ref import MicrobeamManipulationRef
from .objective_settings import ObjectiveSettings
from .pixels import Pixels
from .roi_ref import ROIRef
from .simple_types import ImageID
from .stage_label import StageLabel


class Image(OMEType):
    """This element describes the actual image and its meta-data.

    The elements that are references (ending in Ref or Settings) refer to elements
    defined outside of the Image element. Ref elements are simple links, while
    Settings elements are links with additional values.

    If any of the required Image attributes or elements are missing, its
    guaranteed to be an invalid document. The required attributes and elements are
    ID and Pixels.

    ExperimenterRef is required for all Images with well formed LSIDs. ImageType
    is a vendor-specific designation of the type of image this is. Examples of
    ImageType include 'STK', 'SoftWorx', etc. The Name attributes are in all cases
    the name of the element instance. In this case, the name of the image, not
    necessarily the filename. Physical size of pixels are microns[µm].

    Parameters
    ----------
    id : ImageID
    pixels : Pixels
    acquisition_date : datetime, optional
        The acquisition date of the Image. The element contains an
        xsd:dateTime string based on the ISO 8601 format (i.e.
        1988-04-07T18:39:09.359)  YYYY-MM-DDTHH:mm:SS.sssZ Y - Year M - Month
        D - Day H - Hour m - minutes S - Seconds s - sub-seconds (optional) Z
        - Zone (optional) +HH:mm or -HH:mm or Z for UTC  Note: xsd:dataTime
        supports a very wide date range with unlimited precision. The full
        date range and precision are not typically supported by platform- and
        language-specific libraries. Where the supported time precision is
        less than the precision used by the xsd:dateTime timestamp there will
        be loss of precision; this will typically occur via direct truncation
        or (less commonly) rounding.  The year value can be large and/or
        negative. Any value covering the current or last century should be
        correctly processed, but some systems cannot process earlier dates.
        The sub-second value is defined as an unlimited number of digits after
        the decimal point. In Java a minimum of millisecond precision is
        guaranteed. In C++ microsecond precision is guaranteed, with
        nanosecond precision being available on some platforms.  Time zones
        are supported, eg '2013-10-24T11:52:33+01:00' for Paris, but in most
        cases it will be converted to UTC when the timestamp is written.
    annotation_ref : AnnotationRef, optional
    description : str, optional
        A description for the image.
    experiment_ref : ExperimentRef, optional
    experimenter_group_ref : ExperimenterGroupRef, optional
    experimenter_ref : ExperimenterRef, optional
    imaging_environment : ImagingEnvironment, optional
    instrument_ref : InstrumentRef, optional
    microbeam_manipulation_ref : MicrobeamManipulationRef, optional
    name : str, optional
    objective_settings : ObjectiveSettings, optional
    roi_ref : ROIRef, optional
    stage_label : StageLabel, optional
    """

    id: ImageID
    pixels: Pixels
    acquisition_date: Optional[datetime] = None
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    description: Optional[str] = None
    experiment_ref: Optional[ExperimentRef] = None
    experimenter_group_ref: Optional[ExperimenterGroupRef] = None
    experimenter_ref: Optional[ExperimenterRef] = None
    imaging_environment: Optional[ImagingEnvironment] = None
    instrument_ref: Optional[InstrumentRef] = None
    microbeam_manipulation_ref: List[MicrobeamManipulationRef] = Field(
        default_factory=list
    )
    name: Optional[str] = None
    objective_settings: Optional[ObjectiveSettings] = None
    roi_ref: List[ROIRef] = Field(default_factory=list)
    stage_label: Optional[StageLabel] = None



from typing import Union

from .arc import Arc
from .filament import Filament
from .generic_excitation_source import GenericExcitationSource
from .laser import Laser
from .light_emitting_diode import LightEmittingDiode
from .light_source import LightSource

LightSourceGroup = LightSource

LightSourceGroupType = Union[
    Laser, Arc, Filament, LightEmittingDiode, GenericExcitationSource
]



import pytest

from ome_types.model import OME, Channel, Image, Pixels


@pytest.mark.parametrize("channel_kwargs", [{}, {"color": "blue"}])
def test_color_unset(channel_kwargs: dict) -> None:
    ome = OME(
        images=[
            Image(
                pixels=Pixels(
                    size_c=1,
                    size_t=1,
                    size_x=1,
                    size_y=1,
                    size_z=1,
                    dimension_order="XYZTC",
                    type="uint16",
                    metadata_only=True,
                    channels=[Channel(**channel_kwargs)],
                )
            )
        ]
    )

    assert ("Color" in ome.to_xml()) is bool(channel_kwargs)



from ome_types._base_type import OMEType

from .reference import Reference


class Settings(Reference, OMEType):
    """Settings is an empty complex type that is contained and extended by all the
    `*Settings` elements Each `*Settings` element defines an attribute named ID of
    simple type `*ID` and the other information that is needed.

    Each simple type `*ID` is restricted to the base type LSID with an appropriate
    pattern
    """



from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .experimenter_group_ref import ExperimenterGroupRef
from .experimenter_ref import ExperimenterRef
from .image_ref import ImageRef
from .simple_types import DatasetID


class Dataset(OMEType):
    """An element specifying a collection of images that are always processed
    together.

    Images can belong to more than one Dataset, and a Dataset may contain more
    than one Image. Images contain one or more DatasetRef elements to specify what
    datasets they belong to. Once a Dataset has been processed in any way, its
    collection of images cannot be altered. The ExperimenterRef and
    ExperimenterGroupRef elements specify the person and group this Dataset
    belongs to. Projects may contain one or more Datasets, and Datasets may belong
    to one or more Projects. This relationship is specified by listing DatasetRef
    elements within the Project element.

    Parameters
    ----------
    id : DatasetID
    annotation_ref : AnnotationRef, optional
    description : str, optional
        A description for the dataset.
    experimenter_group_ref : ExperimenterGroupRef, optional
    experimenter_ref : ExperimenterRef, optional
    image_ref : ImageRef, optional
    name : str, optional
        A name for the dataset that is suitable for presentation to the user.
    """

    id: DatasetID
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    description: Optional[str] = None
    experimenter_group_ref: Optional[ExperimenterGroupRef] = None
    experimenter_ref: Optional[ExperimenterRef] = None
    image_ref: List[ImageRef] = Field(default_factory=list)
    name: Optional[str] = None



from typing_extensions import Literal

from ome_types._base_type import OMEType

from .shape import Shape


class Ellipse(Shape, OMEType):
    """A simple ellipse object.

    If rotation is required apply a transformation at the Shape level.

    Parameters
    ----------
    id : ShapeID
    radius_x : float
        The horizontal radius of the ellipse.
    radius_y : float
        The vertical radius of the ellipse.
    x : float
        The X coordinate of the center of the ellipse.
    y : float
        The Y coordinate of the center of the ellipse.
    annotation_ref : AnnotationRef, optional
    fill_color : Color, optional
        The color of the fill - encoded as RGBA The value "-1" is #FFFFFFFF so
        solid white (it is a signed 32 bit value) NOTE: Prior to the 2012-06
        schema the default value was incorrect and produced a transparent red
        not solid white.
    fill_rule : FillRule, optional
    font_family : FontFamily, optional
    font_size : NonNegativeInt, optional
        Size of the font. Units are set by FontSizeUnit.
    font_size_unit : UnitsLength, optional
        The units used for the font size.
    font_style : FontStyle, optional
    locked : bool, optional
        Controls whether the shape is locked and read only, true is locked,
        false is editable.
    stroke_color : Color, optional
        The color of the stroke  - encoded as RGBA The value "-1" is #FFFFFFFF
        so solid white (it is a signed 32 bit value) NOTE: Prior to the
        2012-06 schema the default value was incorrect and produced a
        transparent red not solid white.
    stroke_dash_array : str, optional
        e.g. "none", "10 20 30 10"
    stroke_width : float, optional
        The width of the stroke. Units are set by StrokeWidthUnit.
    stroke_width_unit : UnitsLength, optional
        The units used for the stroke width.
    text : str, optional
    the_c : NonNegativeInt, optional
        The channel the ROI applies to. If not specified then the ROI applies
        to all the channels of the image. This is numbered from 0.
    the_t : NonNegativeInt, optional
        The timepoint the ROI applies to. If not specified then the ROI
        applies to all the timepoints of the image. This is numbered from 0.
    the_z : NonNegativeInt, optional
        The z-section the ROI applies to. If not specified then the ROI
        applies to all the z-sections of the image. This is numbered from 0.
    transform : AffineTransform, optional
        This is a matrix used to transform the shape. The element has 6
        xsd:float attributes. If the element is present then all 6 values must
        be included.
    """

    kind: Literal["ellipse"] = "ellipse"
    radius_x: float
    radius_y: float
    x: float
    y: float



import importlib
import sys
from pathlib import Path

import pytest

ome_autogen = pytest.importorskip("ome_autogen")


def test_autogen(tmp_path_factory):
    """Test that autogen works without raising an exception.

    This does *not* actually test the resulting model.
    """
    target_dir = tmp_path_factory.mktemp("_ome_types_test_model")
    xsd = Path(__file__).parent.parent / "src" / "ome_types" / "ome-2016-06.xsd"
    ome_autogen.convert_schema(url=xsd, target_dir=target_dir)
    sys.path.insert(0, str(target_dir.parent))
    assert importlib.import_module(target_dir.name)
    sys.path.pop(0)



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import ExperimenterID


class ExperimenterRef(Reference, OMEType):
    """This empty element has a required Experimenter ID and an optional DocumentID
    attribute which refers to one of the Experimenters defined within OME.

    Parameters
    ----------
    id : ExperimenterID
    """

    id: ExperimenterID



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import WellSampleID


class WellSampleRef(Reference, OMEType):
    """The WellSampleRef element is a reference to a WellSample element.

    Parameters
    ----------
    id : WellSampleID
    """

    id: WellSampleID



from enum import Enum
from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .manufacturer_spec import ManufacturerSpec
from .simple_types import DetectorID, UnitsElectricPotential


class Type(Enum):
    """The Type of detector.

    E.g. CCD, PMT, EMCCD etc.
    """

    ANALOG_VIDEO = "AnalogVideo"
    APD = "APD"
    CCD = "CCD"
    CMOS = "CMOS"
    CORRELATION_SPECTROSCOPY = "CorrelationSpectroscopy"
    EBCCD = "EBCCD"
    EMCCD = "EMCCD"
    FTIR = "FTIR"
    INTENSIFIED_CCD = "IntensifiedCCD"
    LIFETIME_IMAGING = "LifetimeImaging"
    OTHER = "Other"
    PHOTODIODE = "Photodiode"
    PMT = "PMT"
    SPECTROSCOPY = "Spectroscopy"


class Detector(ManufacturerSpec, OMEType):
    """The type of detector used to capture the image.

    The Detector ID can be used as a reference within the Channel element in the
    Image element. The values stored in Detector represent the fixed values,
    variable values modified during the acquisition go in DetectorSettings

    Each attribute now has an indication of what type of detector it applies to.
    This is preparatory work for cleaning up and possibly splitting this object
    into sub-types.

    Parameters
    ----------
    id : DetectorID
    amplification_gain : float, optional
        Gain applied to the detector signal. This is the electronic gain (as
        apposed to the inherent gain) that is set for the detector.
        {used:EMCCD#EMGain}
    annotation_ref : AnnotationRef, optional
    gain : float, optional
        The Detector Gain for this detector, as a float. {used:CCD,EMCCD,PMT}
    lot_number : str, optional
        The lot number of the component.
    manufacturer : str, optional
        The manufacturer of the component.
    model : str, optional
        The Model of the component.
    offset : float, optional
        The Detector Offset. {used:CCD,EMCCD}
    serial_number : str, optional
        The serial number of the component.
    type : Type, optional
        The Type of detector. E.g. CCD, PMT, EMCCD etc.
    voltage : float, optional
        The Voltage of the detector (e.g. PMT voltage) as a float. {used:PMT}
        Units are set by VoltageUnit.
    voltage_unit : UnitsElectricPotential, optional
        The units of the Voltage - default:volts.
    zoom : float, optional
        The fixed Zoom for a detector. {used:PMT}
    """

    id: DetectorID
    amplification_gain: Optional[float] = None
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    gain: Optional[float] = None
    offset: Optional[float] = None
    type: Optional[Type] = None
    voltage: Optional[float] = None
    voltage_unit: Optional[UnitsElectricPotential] = UnitsElectricPotential("V")
    zoom: Optional[float] = None



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import MicrobeamManipulationID


class MicrobeamManipulationRef(Reference, OMEType):
    """MicrobeamManipulationRef.

    Parameters
    ----------
    id : MicrobeamManipulationID
    """

    id: MicrobeamManipulationID



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import InstrumentID


class InstrumentRef(Reference, OMEType):
    """This empty element can be used (via the required Instrument ID attribute) to
    refer to an Instrument defined within OME.

    Parameters
    ----------
    id : InstrumentID
    """

    id: InstrumentID



from typing import Optional

from ome_types._base_type import OMEType

from .simple_types import UnitsLength


class StageLabel(OMEType):
    """The StageLabel is used to specify a name and position for a stage position in
    the microscope's reference frame.

    Parameters
    ----------
    name : str
    x : float, optional
        The X position of the stage label. Units are set by XUnit.
    x_unit : UnitsLength, optional
        The units of the X stage position - default:.
    y : float, optional
        The Y position of the stage label. Units are set by YUnit.
    y_unit : UnitsLength, optional
        The units of the Y stage position - default:.
    z : float, optional
        The Z position of the stage label. Units are set by ZUnit.
    z_unit : UnitsLength, optional
        The units of the Z  stage position - default:.
    """

    name: str
    x: Optional[float] = None
    x_unit: Optional[UnitsLength] = UnitsLength("reference frame")
    y: Optional[float] = None
    y_unit: Optional[UnitsLength] = UnitsLength("reference frame")
    z: Optional[float] = None
    z_unit: Optional[UnitsLength] = UnitsLength("reference frame")



from datetime import datetime
from typing import Optional

from ome_types._base_type import OMEType

from .image_ref import ImageRef
from .simple_types import NonNegativeInt, UnitsLength, WellSampleID


class WellSample(OMEType):
    """WellSample is an individual image that has been captured within a Well.

    Parameters
    ----------
    id : WellSampleID
    index : NonNegativeInt
        This records the order of the well samples. Each index should be
        unique for a given plate but they do not have to be sequential, there
        may be gaps if part of the dataset is missing. In the user interface
        the displayed value of the index will be calculated modulo the number
        of PlateAcquisitions for the plate.
    image_ref : ImageRef, optional
        This is the main link to the core Image element
    position_x : float, optional
        The X position of the field (image) within the well relative to the
        well origin defined on the Plate. Units are set by PositionXUnit.
    position_x_unit : UnitsLength, optional
        The units of the position in X - default:reference frame.
    position_y : float, optional
        The Y position of the field (image) within the well relative to the
        well origin defined on the Plate. Units are set by PositionYUnit.
    position_y_unit : UnitsLength, optional
        The units of the position in Y - default:reference frame.
    timepoint : datetime, optional
        The time-point at which the image started to be collected
    """

    id: WellSampleID
    index: NonNegativeInt
    image_ref: Optional[ImageRef] = None
    position_x: Optional[float] = None
    position_x_unit: Optional[UnitsLength] = UnitsLength("reference frame")
    position_y: Optional[float] = None
    position_y_unit: Optional[UnitsLength] = UnitsLength("reference frame")
    timepoint: Optional[datetime] = None



from typing_extensions import Literal

from ome_types._base_type import OMEType

from .light_source import LightSource


class LightEmittingDiode(LightSource, OMEType):
    """The LightEmittingDiode element is used to describe various kinds of LED lamps.

    As the LightEmittingDiode is inside a LightSource it already has available the
    values from ManufacturerSpec (Manufacturer, Model, SerialNumber, LotNumber)
    And the values from LightSource which includes Power in milliwatts

    We have looked at extending this element but have had a problem producing a
    generic solution.

    Possible attributes talked about adding include: Power in lumens - but this is
    complicated by multi-channel devices like CoolLED where each channel's power
    is different Wavelength Range - not a simple value so would require multiple
    attributes or a child element Angle of Projection - this would be further
    affected by the optics used for filtering the naked LED or that combine power
    from multiple devices

    These values are further affected if you over-drive the LED resulting in a
    more complex system

    Another issue is that LED's may not be used directly for illumination but as
    drivers for secondary emissions from doped fiber optics. This would require
    the fiber optics to be modeled.

    Thanks to Paul Goodwin of Applied Precision of information about this topic.

    Parameters
    ----------
    id : LightSourceID
        A LightSource ID must be specified for each light source, and the
        individual light sources can be referred to by their LightSource IDs
        (eg from Channel).
    annotation_ref : AnnotationRef, optional
    lot_number : str, optional
        The lot number of the component.
    manufacturer : str, optional
        The manufacturer of the component.
    model : str, optional
        The Model of the component.
    power : float, optional
        The light-source power. Units are set by PowerUnit.
    power_unit : UnitsPower, optional
        The units of the Power - default:milliwatts.
    serial_number : str, optional
        The serial number of the component.
    """

    kind: Literal["light_emitting_diode"] = "light_emitting_diode"



from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .simple_types import ExperimenterID


class Experimenter(OMEType):
    """This element describes a person who performed an imaging experiment.

    This person may also be a user of the OME system, in which case the UserName
    element contains their login name. Experimenters may belong to one or more
    groups which are specified using one or more ExperimenterGroupRef elements.

    Parameters
    ----------
    id : ExperimenterID
    annotation_ref : AnnotationRef, optional
    email : str, optional
        A person's email address.
    first_name : str, optional
        First name, sometime called christian name or given name or forename.
        [plain text string]
    institution : str, optional
        A person's Institution The organizing structure that people belong to
        other than groups.  A university, or company, etc. We do not specify a
        department element, and do not mean for Institution to be used in this
        way. We simply wish to say XXX at YYY.  Where YYY has a better chance
        of being tied to a geographically fixed location and of being more
        recognizable than a group of experimenters.
    last_name : str, optional
        A person's last name sometimes called surname or family name. [plain
        text string]
    middle_name : str, optional
        Any other names.
    user_name : str, optional
        This is the username of the experimenter (in a 'unix' or 'database'
        sense).
    """

    id: ExperimenterID
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    email: Optional[str] = None
    first_name: Optional[str] = None
    institution: Optional[str] = None
    last_name: Optional[str] = None
    middle_name: Optional[str] = None
    user_name: Optional[str] = None



from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .experimenter_ref import ExperimenterRef
from .leader import Leader
from .simple_types import ExperimenterGroupID


class ExperimenterGroup(OMEType):
    """The ExperimenterGroupID is required.

    Information should ideally be specified for at least one Leader as a contact
    for the group. The Leaders are themselves Experimenters.

    Parameters
    ----------
    id : ExperimenterGroupID
    annotation_ref : AnnotationRef, optional
    description : str, optional
        A description for the group.
    experimenter_ref : ExperimenterRef, optional
    leader : Leader, optional
    name : str, optional
    """

    id: ExperimenterGroupID
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    description: Optional[str] = None
    experimenter_ref: List[ExperimenterRef] = Field(default_factory=list)
    leader: List[Leader] = Field(default_factory=list)
    name: Optional[str] = None



from typing import List

from pydantic import Field

from ome_types._base_type import OMEType


class M(OMEType):
    """This is a key/value pair used to build up a Mapping.

    The          Element and Attribute name are kept to single letters to minimize
    the          length at the expense of readability as they are likely to occur
    many          times.

    Parameters
    ----------
    k : str, optional
    """

    value: str
    k: str


class Map(OMEType):
    """This is a Mapping of key/value pairs.

    Parameters
    ----------
    k : str, optional
    m : M, optional
        This is a key/value pair used to build up a Mapping. The
        Element and Attribute name are kept to single letters to minimize the
        length at the expense of readability as they are likely to occur many
        times.
    """

    m: List[M] = Field(default_factory=list)



from typing import Optional

from typing_extensions import Literal

from ome_types._base_type import OMEType

from .light_source import LightSource
from .map import Map


class GenericExcitationSource(LightSource, OMEType):
    """The GenericExcitationSource element is used to represent a source as a
    collection of key/value pairs, stored in a Map.

    The other lightsource objects should always be used in preference to this if
    possible.

    Parameters
    ----------
    id : LightSourceID
        A LightSource ID must be specified for each light source, and the
        individual light sources can be referred to by their LightSource IDs
        (eg from Channel).
    annotation_ref : AnnotationRef, optional
    lot_number : str, optional
        The lot number of the component.
    manufacturer : str, optional
        The manufacturer of the component.
    map : Map, optional
    model : str, optional
        The Model of the component.
    power : float, optional
        The light-source power. Units are set by PowerUnit.
    power_unit : UnitsPower, optional
        The units of the Power - default:milliwatts.
    serial_number : str, optional
        The serial number of the component.
    """

    kind: Literal["generic_excitation_source"] = "generic_excitation_source"
    map: Optional[Map] = None



from typing import Any, Callable, Dict, Generator, List, Optional
from xml.etree import ElementTree

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .text_annotation import TextAnnotation


class Element(ElementTree.Element):
    """ElementTree.Element that supports pydantic validation."""

    @classmethod
    def __get_validators__(cls) -> Generator[Callable[[Any], Any], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, v: Any) -> ElementTree.Element:
        if isinstance(v, ElementTree.Element):
            return v
        try:
            return ElementTree.fromstring(v)
        except ElementTree.ParseError as e:
            raise ValueError(f"Invalid XML string: {e}")


class XMLAnnotation(TextAnnotation, OMEType):
    """An general xml annotation.

    The contents of this is not processed as OME XML but should still be well-
    formed XML.

    Parameters
    ----------
    id : AnnotationID
    value : Element
    annotation_ref : AnnotationRef, optional
    annotator : ExperimenterID, optional
        The Annotator is the person who attached this annotation. e.g. If
        UserA annotates something with TagB, owned by UserB, UserA is still
        the Annotator.
    description : str, optional
        A description for the annotation.
    namespace : str, optional
        We recommend the inclusion of a namespace for annotations you define.
        If it is absent then we assume the annotation is to use our (OME's)
        default interpretation for this type.
    """

    value: Element
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    description: Optional[str] = None

    # NOTE: pickling this object requires xmlschema>=1.4.1

    def dict(self, **k: Any) -> Dict[str, Any]:
        d = super().dict(**k)
        d["value"] = ElementTree.tostring(
            d.pop("value"), encoding="unicode", method="xml"
        ).strip()
        return d



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import LightSourceID


class Pump(Reference, OMEType):
    """The Pump element is a reference to a LightSource.

    It is used within the Laser element to specify the light source for the
    laser's pump (if any).

    Parameters
    ----------
    id : LightSourceID
    """

    id: LightSourceID



from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .plate_acquisition import PlateAcquisition
from .simple_types import (
    NamingConvention,
    NonNegativeInt,
    PlateID,
    PositiveInt,
    UnitsLength,
)
from .well import Well


class Plate(OMEType):
    """This element identifies microtiter plates within a screen.

    A plate can belong to more than one screen. The Screen(s) that a plate belongs
    to are specified by the ScreenRef element. The Plate ID and Name attributes
    are required. The Wells in a plate are numbers from the top-left corner in a
    grid starting at zero. i.e. The top-left well of a plate is index (0,0)

    Parameters
    ----------
    id : PlateID
    annotation_ref : AnnotationRef, optional
    column_naming_convention : NamingConvention, optional
        The ColumnNamingConvention
    columns : PositiveInt, optional
        The number of columns in the plate
    description : str, optional
        A description for the plate.
    external_identifier : str, optional
        The ExternalIdentifier attribute may contain a reference to an
        external database.
    field_index : NonNegativeInt, optional
        The index of the WellSample to display as the default Field
    name : str, optional
        The Name identifies the plate to the user. It is used much like the
        ID, and so must be unique within the document.  If a plate name is not
        available when one is needed it will be constructed in the following
        order: 1. If name is available use it. 2. If not use "Start time - End
        time" (NOTE: Not a subtraction! A string representation of the two
        times separated by a dash.) 3. If these times are not available use
        the Plate ID.
    plate_acquisitions : PlateAcquisition, optional
    row_naming_convention : NamingConvention, optional
        The RowNamingConvention
    rows : PositiveInt, optional
        The number of rows in the plate
    status : str, optional
        A textual annotation of the current state of the plate with respect to
        the experiment work-flow; e.g. 1. Seed cell: done; 2. Transfection:
        done;      3. Gel doc: todo.
    well_origin_x : float, optional
        This defines the X position to use for the origin of the fields
        (individual images) taken in a well. It is used with the X in the
        WellSample to display the fields in the correct position relative to
        each other. Each Well in the plate has the same well origin. Units are
        set by WellOriginXUnit.  In the OMERO clients by convention we display
        the WellOrigin in the center of the view.
    well_origin_x_unit : UnitsLength, optional
        The units of the well origin in X - default:reference frame.
    well_origin_y : float, optional
        This defines the Y position to use for the origin of the fields
        (individual images) taken in a well. It is used with the Y in the
        WellSample to display the fields in the correct position relative to
        each other. Each Well in the plate has the same well origin.  Units
        are set by WellOriginYUnit.  In the OMERO clients by convention we
        display the WellOrigin in the center of the view.
    well_origin_y_unit : UnitsLength, optional
        The units of the well origin in Y - default:reference frame.
    wells : Well, optional
    """

    id: PlateID
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    column_naming_convention: Optional[NamingConvention] = None
    columns: Optional[PositiveInt] = None
    description: Optional[str] = None
    external_identifier: Optional[str] = None
    field_index: Optional[NonNegativeInt] = None
    name: Optional[str] = None
    plate_acquisitions: List[PlateAcquisition] = Field(default_factory=list)
    row_naming_convention: Optional[NamingConvention] = None
    rows: Optional[PositiveInt] = None
    status: Optional[str] = None
    well_origin_x: Optional[float] = None
    well_origin_x_unit: Optional[UnitsLength] = UnitsLength("reference frame")
    well_origin_y: Optional[float] = None
    well_origin_y_unit: Optional[UnitsLength] = UnitsLength("reference frame")
    wells: List[Well] = Field(default_factory=list)



from typing import Optional

from ome_types._base_type import OMEType

from .settings import Settings
from .simple_types import LightSourceID, PercentFraction, PositiveFloat, UnitsLength


class LightSourceSettings(Settings, OMEType):
    """LightSourceSettings.

    Parameters
    ----------
    id : LightSourceID
    attenuation : PercentFraction, optional
        The Attenuation of the light source A fraction, as a value from 0.0 to
        1.0.
    wavelength : PositiveFloat, optional
        The Wavelength of the light source. Units are set by WavelengthUnit.
    wavelength_unit : UnitsLength, optional
        The units of the Wavelength of the light source - default:nanometres
    """

    id: LightSourceID
    attenuation: Optional[PercentFraction] = None
    wavelength: Optional[PositiveFloat] = None
    wavelength_unit: Optional[UnitsLength] = UnitsLength("nm")



from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .dataset_ref import DatasetRef
from .experimenter_group_ref import ExperimenterGroupRef
from .experimenter_ref import ExperimenterRef
from .simple_types import ProjectID


class Project(OMEType):
    """The Project ID is required.

    Datasets can be grouped into projects using a many-to-many relationship. A
    Dataset may belong to one or more Projects by including one or more ProjectRef
    elements which refer to Project IDs. Projects do not directly contain images -
    only by virtue of containing datasets, which themselves contain images.

    Parameters
    ----------
    id : ProjectID
    annotation_ref : AnnotationRef, optional
    dataset_ref : DatasetRef, optional
    description : str, optional
        A description for the project.
    experimenter_group_ref : ExperimenterGroupRef, optional
    experimenter_ref : ExperimenterRef, optional
    name : str, optional
    """

    id: ProjectID
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    dataset_ref: List[DatasetRef] = Field(default_factory=list)
    description: Optional[str] = None
    experimenter_group_ref: Optional[ExperimenterGroupRef] = None
    experimenter_ref: Optional[ExperimenterRef] = None
    name: Optional[str] = None



from datetime import datetime
from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .basic_annotation import BasicAnnotation


class TimestampAnnotation(BasicAnnotation, OMEType):
    """A date/time annotation of type xsd:dateTime

    Parameters
    ----------
    id : AnnotationID
    value : datetime
    annotation_ref : AnnotationRef, optional
    annotator : ExperimenterID, optional
        The Annotator is the person who attached this annotation. e.g. If
        UserA annotates something with TagB, owned by UserB, UserA is still
        the Annotator.
    description : str, optional
        A description for the annotation.
    namespace : str, optional
        We recommend the inclusion of a namespace for annotations you define.
        If it is absent then we assume the annotation is to use our (OME's)
        default interpretation for this type.
    """

    value: datetime
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    description: Optional[str] = None



from typing import Optional

from ome_types._base_type import OMEType

from .bin_data import BinData
from .external import External
from .simple_types import NonNegativeLong


class BinaryFile(OMEType):
    """Describes a binary file.

    Parameters
    ----------
    file_name : str
    size : NonNegativeLong
        Size of the uncompressed file.
    bin_data : BinData, optional
    external : External, optional
    mime_type : str, optional
    """

    file_name: str
    size: NonNegativeLong
    bin_data: Optional[BinData] = None
    external: Optional[External] = None
    mime_type: Optional[str] = None



from typing import List

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .manufacturer_spec import ManufacturerSpec
from .simple_types import DichroicID


class Dichroic(ManufacturerSpec, OMEType):
    """The dichromatic beamsplitter or dichroic mirror used for this filter
    combination.

    Parameters
    ----------
    id : DichroicID
    annotation_ref : AnnotationRef, optional
    lot_number : str, optional
        The lot number of the component.
    manufacturer : str, optional
        The manufacturer of the component.
    model : str, optional
        The Model of the component.
    serial_number : str, optional
        The serial number of the component.
    """

    id: DichroicID
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)



from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .dichroic_ref import DichroicRef
from .filter_ref import FilterRef


class LightPath(OMEType):
    """A description of the light path

    Parameters
    ----------
    annotation_ref : AnnotationRef, optional
    dichroic_ref : DichroicRef, optional
    emission_filter_ref : FilterRef, optional
        The Filters placed in the Emission light path.
    excitation_filter_ref : FilterRef, optional
        The Filters placed in the Excitation light path.
    """

    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    dichroic_ref: Optional[DichroicRef] = None
    emission_filter_ref: List[FilterRef] = Field(default_factory=list)
    excitation_filter_ref: List[FilterRef] = Field(default_factory=list)



import pickle
import re
from pathlib import Path
from unittest import mock
from xml.dom import minidom
from xml.etree import ElementTree

import pytest
from pydantic import ValidationError
from xmlschema.validators.exceptions import XMLSchemaValidationError

import util
from ome_types import from_tiff, from_xml, model, to_xml
from ome_types._xmlschema import NS_OME, URI_OME, get_schema, to_xml_element

ValidationErrors = [ValidationError, XMLSchemaValidationError]
try:
    from lxml.etree import XMLSchemaValidateError

    ValidationErrors.append(XMLSchemaValidateError)
except ImportError:
    pass


SHOULD_FAIL_READ = {
    # Some timestamps have negative years which datetime doesn't support.
    "timestampannotation",
}
SHOULD_FAIL_VALIDATION = {"invalid_xml_annotation"}
SHOULD_RAISE_READ = {"bad"}
SHOULD_FAIL_ROUNDTRIP = {
    # Order of elements in StructuredAnnotations and Union are jumbled.
    "timestampannotation-posix-only",
    "transformations-downgrade",
    "invalid_xml_annotation",
}
SHOULD_FAIL_ROUNDTRIP_LXML = {
    "folders-simple-taxonomy",
    "folders-larger-taxonomy",
}
SKIP_ROUNDTRIP = {
    # These have XMLAnnotations with extra namespaces and mixed content, which
    # the automated round-trip test code doesn't properly verify yet. So even
    # though these files do appear to round-trip correctly when checked by eye,
    # we'll play it safe and skip them until the test is fixed.
    "spim",
    "xmlannotation-body-space",
    "xmlannotation-multi-value",
    "xmlannotation-svg",
}


def mark_xfail(fname):
    return pytest.param(
        fname,
        marks=pytest.mark.xfail(
            strict=True, reason="Unexpected success. You fixed it!"
        ),
    )


def mark_skip(fname):
    return pytest.param(fname, marks=pytest.mark.skip)


def true_stem(p):
    return p.name.partition(".")[0]


all_xml = list((Path(__file__).parent / "data").glob("*.ome.xml"))
xml_read = [mark_xfail(f) if true_stem(f) in SHOULD_FAIL_READ else f for f in all_xml]
xml_roundtrip = []
for f in all_xml:
    stem = true_stem(f)
    if stem in SHOULD_FAIL_READ | SHOULD_RAISE_READ:
        continue
    elif stem in SHOULD_FAIL_ROUNDTRIP:
        f = mark_xfail(f)
    elif stem in SKIP_ROUNDTRIP:
        f = mark_skip(f)
    xml_roundtrip.append(f)


validate = [True, False]

parser = ["lxml", "xmlschema"]


@pytest.mark.parametrize("xml", xml_read, ids=true_stem)
@pytest.mark.parametrize("parser", parser)
@pytest.mark.parametrize("validate", validate)
def test_from_xml(xml, parser: str, validate: bool, benchmark):
    should_raise = SHOULD_RAISE_READ.union(SHOULD_FAIL_VALIDATION if validate else [])
    if true_stem(xml) in should_raise:
        with pytest.raises(tuple(ValidationErrors)):
            assert benchmark(from_xml, xml, parser=parser, validate=validate)
    else:
        assert benchmark(from_xml, xml, parser=parser, validate=validate)


@pytest.mark.parametrize("parser", parser)
@pytest.mark.parametrize("validate", validate)
def test_from_tiff(benchmark, validate, parser):
    """Test that OME metadata extractions from Tiff headers works."""
    _path = Path(__file__).parent / "data" / "ome.tiff"
    ome = benchmark(from_tiff, _path, parser=parser, validate=validate)
    assert len(ome.images) == 1
    assert ome.images[0].id == "Image:0"
    assert ome.images[0].pixels.size_x == 6
    assert ome.images[0].pixels.channels[0].samples_per_pixel == 1


@pytest.mark.parametrize("xml", xml_roundtrip, ids=true_stem)
@pytest.mark.parametrize("parser", parser)
@pytest.mark.parametrize("validate", validate)
def test_roundtrip(xml, parser, validate, benchmark):
    """Ensure we can losslessly round-trip XML through the model and back."""
    xml = str(xml)
    schema = get_schema(xml)

    def canonicalize(xml, strip_empty):
        d = schema.decode(xml, use_defaults=True)
        # Strip extra whitespace in the schemaLocation value.
        d["@xsi:schemaLocation"] = re.sub(r"\s+", " ", d["@xsi:schemaLocation"])
        root = schema.encode(d, path=NS_OME + "OME", use_defaults=True)
        # These are the tags that appear in the example files with empty
        # content. Since our round-trip will drop empty elements, we'll need to
        # strip them from the "original" documents before comparison.
        if strip_empty:
            for tag in ("Description", "LightPath", "Map"):
                for e in root.findall(f".//{NS_OME}{tag}[.='']..."):
                    e.remove(e.find(f"{NS_OME}{tag}"))
        # ET.canonicalize can't handle an empty namespace so we need to
        # re-register the OME namespace with an actual name before calling
        # tostring.
        ElementTree.register_namespace("ome", URI_OME)
        xml_out = ElementTree.tostring(root, "unicode")
        xml_out = util.canonicalize(xml_out, strip_text=True)
        xml_out = minidom.parseString(xml_out).toprettyxml(indent="  ")
        return xml_out

    original = canonicalize(xml, True)
    ome = from_xml(xml, parser=parser, validate=validate)
    rexml = benchmark(to_xml, ome)

    try:
        assert canonicalize(rexml, False) == original
    except AssertionError:
        # Special xfail catch since two files fail only with xml2dict
        if true_stem(Path(xml)) in SHOULD_FAIL_ROUNDTRIP_LXML and parser == "lxml":
            pytest.xfail(
                f"Expected failure on roundtrip using xml2dict on file: {stem}"
            )
        else:
            raise


@pytest.mark.parametrize("parser", parser)
@pytest.mark.parametrize("validate", validate)
def test_to_xml_with_kwargs(validate, parser):
    """Ensure kwargs are passed to ElementTree"""
    ome = from_xml(
        Path(__file__).parent / "data" / "example.ome.xml",
        parser=parser,
        validate=validate,
    )

    with mock.patch("xml.etree.ElementTree.tostring") as mocked_et_tostring:
        element = to_xml_element(ome)
        # Use an ElementTree.tostring kwarg and assert that it was passed through
        to_xml(element, xml_declaration=True)
        assert mocked_et_tostring.call_args.xml_declaration


@pytest.mark.parametrize("xml", xml_read, ids=true_stem)
@pytest.mark.parametrize("parser", parser)
@pytest.mark.parametrize("validate", validate)
def test_serialization(xml, validate, parser):
    """Test pickle serialization and reserialization."""
    if true_stem(xml) in SHOULD_RAISE_READ:
        pytest.skip("Can't pickle unreadable xml")
    if validate and true_stem(xml) in SHOULD_FAIL_VALIDATION:
        pytest.skip("Can't pickle invalid xml with validate=True")

    ome = from_xml(xml, parser=parser, validate=validate)
    serialized = pickle.dumps(ome)
    deserialized = pickle.loads(serialized)
    assert ome == deserialized


def test_no_id():
    """Test that ids are optional, and auto-increment."""
    i = model.Instrument(id=20)
    assert i.id == "Instrument:20"
    i2 = model.Instrument()
    assert i2.id == "Instrument:21"

    # but validation still works
    with pytest.raises(ValueError):
        model.Instrument(id="nonsense")


def test_required_missing():
    """Test subclasses with non-default arguments still work."""
    with pytest.raises(ValidationError) as e:
        _ = model.BooleanAnnotation()
    assert "1 validation error for BooleanAnnotation" in str(e.value)
    assert "value\n  field required" in str(e.value)

    with pytest.raises(ValidationError) as e:
        _ = model.Label()
    assert "2 validation errors for Label" in str(e.value)
    assert "x\n  field required" in str(e.value)
    assert "y\n  field required" in str(e.value)


@pytest.mark.parametrize("parser", parser)
@pytest.mark.parametrize("validate", validate)
def test_refs(validate, parser):
    xml = Path(__file__).parent / "data" / "two-screens-two-plates-four-wells.ome.xml"
    ome = from_xml(xml, parser=parser, validate=validate)
    assert ome.screens[0].plate_ref[0].ref is ome.plates[0]


@pytest.mark.parametrize("validate", validate)
@pytest.mark.parametrize("parser", parser)
def test_with_ome_ns(validate, parser):
    xml = Path(__file__).parent / "data" / "ome_ns.ome.xml"
    ome = from_xml(xml, parser=parser, validate=validate)
    assert ome.experimenters



from ome_types._base_type import OMEType

from .annotation import Annotation


class ListAnnotation(Annotation, OMEType):
    """This annotation is a grouping object.

    It uses the sequence of annotation refs from the base Annotation to form the
    list.

    Parameters
    ----------
    id : AnnotationID
    annotation_ref : AnnotationRef, optional
    annotator : ExperimenterID, optional
        The Annotator is the person who attached this annotation. e.g. If
        UserA annotates something with TagB, owned by UserB, UserA is still
        the Annotator.
    description : str, optional
        A description for the annotation.
    namespace : str, optional
        We recommend the inclusion of a namespace for annotations you define.
        If it is absent then we assume the annotation is to use our (OME's)
        default interpretation for this type.
    """



from pathlib import Path

import pytest

from ome_types import from_xml

DATA = Path(__file__).parent / "data"


def test_bad_xml_annotation() -> None:
    """Test that a bad XML annotation is not imported"""
    with pytest.warns(match="Casting invalid AnnotationID"):
        ome = from_xml(DATA / "invalid_xml_annotation.ome.xml")
    assert len(ome.images) == 1
    assert ome.structured_annotations[0].id == "Annotation:0"



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import FilterID


class FilterRef(Reference, OMEType):
    """FilterRef.

    Parameters
    ----------
    id : FilterID
    """

    id: FilterID



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import ExperimenterID


class Leader(Reference, OMEType):
    """Contact information for a ExperimenterGroup leader specified using a reference
    to an Experimenter element defined elsewhere in the document.

    Parameters
    ----------
    id : ExperimenterID
    """

    id: ExperimenterID



from typing import Optional

from ome_types._base_type import OMEType

from .map import Map
from .simple_types import PercentFraction, UnitsPressure, UnitsTemperature


class ImagingEnvironment(OMEType):
    """This describes the environment that the biological sample was in during the
    experiment.

    Parameters
    ----------
    air_pressure : float, optional
        AirPressure is the define units.
    air_pressure_unit : UnitsPressure, optional
        The units the AirPressure is in - default:millibars.
    co2_percent : PercentFraction, optional
        Carbon Dioxide concentration around the sample A fraction, as a value
        from 0.0 to 1.0.
    humidity : PercentFraction, optional
        Humidity around the sample A fraction, as a value from 0.0 to 1.0.
    map : Map, optional
    temperature : float, optional
        The Temperature is the define units.
    temperature_unit : UnitsTemperature, optional
        The units the Temperature is in - default:Celsius.
    """

    air_pressure: Optional[float] = None
    air_pressure_unit: Optional[UnitsPressure] = UnitsPressure("mbar")
    co2_percent: Optional[PercentFraction] = None
    humidity: Optional[PercentFraction] = None
    map: Optional[Map] = None
    temperature: Optional[float] = None
    temperature_unit: Optional[UnitsTemperature] = UnitsTemperature("°C")



from typing import Any, Dict, Iterator, List, Optional, Sequence

from pydantic import Field, validator

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .shape_group import ShapeGroupType
from .simple_types import ROIID


class ROI(OMEType):
    """A four dimensional 'Region of Interest'.

    If they are not used, and the Image has more than one plane, the entire set of
    planes is assumed to be included in the ROI. Multiple ROIs may be specified.

    Parameters
    ----------
    id : ROIID
    annotation_ref : AnnotationRef, optional
    description : str, optional
        A description for the ROI.
    name : str, optional
        The Name identifies the ROI to the user.
    union : List[ShapeGroupType], optional
    """

    id: ROIID
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    description: Optional[str] = None
    name: Optional[str] = None
    union: List[ShapeGroupType] = Field(default_factory=list)

    @validator("union", pre=True)
    def _validate_union(cls, value: Any) -> Sequence[Dict[str, Any]]:
        if isinstance(value, dict):
            return list(cls._flatten_union_dict(value))
        if not isinstance(value, Sequence):
            raise TypeError("must be dict or sequence of dicts")
        return value

    @classmethod
    def _flatten_union_dict(
        cls, nested: Dict[str, Any], keyname: str = "kind"
    ) -> Iterator[Dict[str, Any]]:
        for key, value in nested.items():
            keydict = {keyname: key} if keyname else {}
            if isinstance(value, list):
                yield from ({**x, **keydict} for x in value)
            else:
                yield {**value, **keydict}



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import ROIID


class ROIRef(Reference, OMEType):
    """ROIRef.

    Parameters
    ----------
    id : ROIID
    """

    id: ROIID



from enum import Enum
from typing import Optional

from ome_types._base_type import OMEType

from .manufacturer_spec import ManufacturerSpec


class Type(Enum):
    DISSECTION = "Dissection"
    ELECTROPHYSIOLOGY = "Electrophysiology"
    INVERTED = "Inverted"
    OTHER = "Other"
    UPRIGHT = "Upright"


class Microscope(ManufacturerSpec, OMEType):
    """The microscope's manufacturer specification.

    Parameters
    ----------
    lot_number : str, optional
        The lot number of the component.
    manufacturer : str, optional
        The manufacturer of the component.
    model : str, optional
        The Model of the component.
    serial_number : str, optional
        The serial number of the component.
    type : Type, optional
    """

    type: Optional[Type] = None



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import ExperimentID


class ExperimentRef(Reference, OMEType):
    """ExperimentRef.

    Parameters
    ----------
    id : ExperimentID
    """

    id: ExperimentID



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import ImageID


class ImageRef(Reference, OMEType):
    """The ImageRef element is a reference to an Image element.

    Parameters
    ----------
    id : ImageID
    """

    id: ImageID



import weakref
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from pydantic import Field, validator

from ome_types import util
from ome_types._base_type import OMEType

from .annotation import Annotation
from .boolean_annotation import BooleanAnnotation
from .comment_annotation import CommentAnnotation
from .dataset import Dataset
from .double_annotation import DoubleAnnotation
from .experiment import Experiment
from .experimenter import Experimenter
from .experimenter_group import ExperimenterGroup
from .file_annotation import FileAnnotation
from .folder import Folder
from .image import Image
from .instrument import Instrument
from .list_annotation import ListAnnotation
from .long_annotation import LongAnnotation
from .map_annotation import MapAnnotation
from .plate import Plate
from .project import Project
from .rights import Rights
from .roi import ROI
from .screen import Screen
from .simple_types import UniversallyUniqueIdentifier
from .tag_annotation import TagAnnotation
from .term_annotation import TermAnnotation
from .timestamp_annotation import TimestampAnnotation
from .xml_annotation import XMLAnnotation

_annotation_types: Dict[str, type] = {
    "boolean_annotation": BooleanAnnotation,
    "comment_annotation": CommentAnnotation,
    "double_annotation": DoubleAnnotation,
    "file_annotation": FileAnnotation,
    "list_annotation": ListAnnotation,
    "long_annotation": LongAnnotation,
    "map_annotation": MapAnnotation,
    "tag_annotation": TagAnnotation,
    "term_annotation": TermAnnotation,
    "timestamp_annotation": TimestampAnnotation,
    "xml_annotation": XMLAnnotation,
}


class BinaryOnly(OMEType):
    """Pointer to an external metadata file.

    If this              element is present, then no other metadata may be present
    in this              file, i.e. this file is a place-holder.

    Parameters
    ----------
    metadata_file : str
        Filename of the OME-XML metadata file for                  this binary
        data. If the file cannot be found, a search can                  be
        performed based on the UUID.
    uuid : UniversallyUniqueIdentifier
        The unique identifier of another OME-XML                  block whose
        metadata describes the binary data in this file.                  This
        UUID is considered authoritative regardless of
        mismatches in the filename.
    """

    metadata_file: str
    uuid: UniversallyUniqueIdentifier


class OME(OMEType):
    """The OME element is a container for all information objects accessible by OME.

    These information objects include descriptions of the imaging experiments and
    the people who perform them, descriptions of the microscope, the resulting
    images and how they were acquired, the analyses performed on those images, and
    the analysis results themselves. An OME file may contain any or all of this
    information.

    With the creation of the Metadata Only Companion OME-XML and Binary Only OME-
    TIFF files the top level OME node has changed slightly. It can EITHER: Contain
    all the previously expected elements OR: Contain a single BinaryOnly element
    that points at its Metadata Only Companion OME-XML file.

    Parameters
    ----------
    binary_only : BinaryOnly, optional
        Pointer to an external metadata file. If this              element is
        present, then no other metadata may be present in this
        file, i.e. this file is a place-holder.
    creator : str, optional
        This is the name of the creating application of the OME-XML and
        preferably its full version. e.g "CompanyName, SoftwareName,
        V2.6.3456" This is optional but we hope it will be set by applications
        writing out OME-XML from scratch.
    datasets : Dataset, optional
    experimenter_groups : ExperimenterGroup, optional
    experimenters : Experimenter, optional
    experiments : Experiment, optional
    folders : Folder, optional
    images : Image, optional
    instruments : Instrument, optional
    plates : Plate, optional
    projects : Project, optional
    rights : Rights, optional
    rois : ROI, optional
    screens : Screen, optional
    structured_annotations : List[Annotation], optional
    uuid : UniversallyUniqueIdentifier, optional
        This unique identifier is used to keep track of multi part files. It
        allows the links between files to survive renaming.  While OPTIONAL in
        the general case this is REQUIRED in a MetadataOnly Companion to a
        collection of BinaryOnly files.
    """

    binary_only: Optional[BinaryOnly] = None
    creator: Optional[str] = None
    datasets: List[Dataset] = Field(default_factory=list)
    experimenter_groups: List[ExperimenterGroup] = Field(default_factory=list)
    experimenters: List[Experimenter] = Field(default_factory=list)
    experiments: List[Experiment] = Field(default_factory=list)
    folders: List[Folder] = Field(default_factory=list)
    images: List[Image] = Field(default_factory=list)
    instruments: List[Instrument] = Field(default_factory=list)
    plates: List[Plate] = Field(default_factory=list)
    projects: List[Project] = Field(default_factory=list)
    rights: Optional[Rights] = None
    rois: List[ROI] = Field(default_factory=list)
    screens: List[Screen] = Field(default_factory=list)
    structured_annotations: List[Annotation] = Field(default_factory=list)
    uuid: Optional[UniversallyUniqueIdentifier] = None

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self._link_refs()

    def _link_refs(self) -> None:
        ids = util.collect_ids(self)
        for ref in util.collect_references(self):
            ref._ref = weakref.ref(ids[ref.id])

    def __setstate__(self: Any, state: Dict[str, Any]) -> None:
        """Support unpickle of our weakref references."""
        super().__setstate__(state)
        self._link_refs()

    @classmethod
    def from_xml(cls, xml: Union[Path, str]) -> "OME":
        from ome_types import from_xml

        return from_xml(xml)

    @classmethod
    def from_tiff(cls, path: Union[Path, str]) -> "OME":
        from ome_types import from_tiff

        return from_tiff(path)

    def to_xml(self) -> str:
        from ome_types import to_xml

        return to_xml(self)

    @validator("structured_annotations", pre=True, each_item=True)
    def validate_structured_annotations(
        cls, value: Union[Annotation, Dict[Any, Any]]
    ) -> Annotation:
        if isinstance(value, Annotation):
            return value
        elif isinstance(value, dict):
            try:
                _type = value.pop("_type")
            except KeyError:
                raise ValueError("dict initialization requires _type") from None
            try:
                annotation_cls = _annotation_types[_type]
            except KeyError:
                raise ValueError(f"unknown Annotation type '{_type}'") from None
            return annotation_cls(**value)
        else:
            raise ValueError("invalid type for annotation values")



from enum import Enum
from typing import Optional

from typing_extensions import Literal

from ome_types._base_type import OMEType

from .light_source import LightSource


class Type(Enum):
    """The type of filament."""

    HALOGEN = "Halogen"
    INCANDESCENT = "Incandescent"
    OTHER = "Other"


class Filament(LightSource, OMEType):
    """The Filament element is used to describe various kinds of filament bulbs such
    as Incadescent or Halogen.

    The Power of the Filament is now stored in the LightSource.

    Parameters
    ----------
    id : LightSourceID
        A LightSource ID must be specified for each light source, and the
        individual light sources can be referred to by their LightSource IDs
        (eg from Channel).
    annotation_ref : AnnotationRef, optional
    lot_number : str, optional
        The lot number of the component.
    manufacturer : str, optional
        The manufacturer of the component.
    model : str, optional
        The Model of the component.
    power : float, optional
        The light-source power. Units are set by PowerUnit.
    power_unit : UnitsPower, optional
        The units of the Power - default:milliwatts.
    serial_number : str, optional
        The serial number of the component.
    type : Type, optional
        The type of filament.
    """

    kind: Literal["filament"] = "filament"
    type: Optional[Type] = None



from typing import Optional

from ome_types._base_type import OMEType

from .boolean_annotation import BooleanAnnotation
from .comment_annotation import CommentAnnotation
from .double_annotation import DoubleAnnotation
from .file_annotation import FileAnnotation
from .list_annotation import ListAnnotation
from .long_annotation import LongAnnotation
from .map_annotation import MapAnnotation
from .tag_annotation import TagAnnotation
from .term_annotation import TermAnnotation
from .timestamp_annotation import TimestampAnnotation
from .xml_annotation import XMLAnnotation


class StructuredAnnotations(OMEType):
    """An unordered collection of annotation attached to objects in the OME data
    model.

    Parameters
    ----------
    boolean_annotations : BooleanAnnotation, optional
    comment_annotations : CommentAnnotation, optional
    double_annotations : DoubleAnnotation, optional
    file_annotations : FileAnnotation, optional
    list_annotations : ListAnnotation, optional
    long_annotations : LongAnnotation, optional
    map_annotations : MapAnnotation, optional
    tag_annotations : TagAnnotation, optional
    term_annotations : TermAnnotation, optional
    timestamp_annotations : TimestampAnnotation, optional
    xml_annotations : XMLAnnotation, optional
    """

    boolean_annotations: Optional[BooleanAnnotation] = None
    comment_annotations: Optional[CommentAnnotation] = None
    double_annotations: Optional[DoubleAnnotation] = None
    file_annotations: Optional[FileAnnotation] = None
    list_annotations: Optional[ListAnnotation] = None
    long_annotations: Optional[LongAnnotation] = None
    map_annotations: Optional[MapAnnotation] = None
    tag_annotations: Optional[TagAnnotation] = None
    term_annotations: Optional[TermAnnotation] = None
    timestamp_annotations: Optional[TimestampAnnotation] = None
    xml_annotations: Optional[XMLAnnotation] = None



from datetime import datetime
from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .simple_types import PlateAcquisitionID, PositiveInt
from .well_sample_ref import WellSampleRef


class PlateAcquisition(OMEType):
    """PlateAcquisition is used to describe a single acquisition run for a plate.

    This object is used to record the set of images acquired in a single
    acquisition run. The Images for this run are linked to PlateAcquisition
    through WellSample.

    Parameters
    ----------
    id : PlateAcquisitionID
    annotation_ref : AnnotationRef, optional
    description : str, optional
        A description for the PlateAcquisition.
    end_time : datetime, optional
        Time when the last image of this acquisition was collected
    maximum_field_count : PositiveInt, optional
        The maximum number of fields (well samples) in any well in this
        PlateAcquisition. This is only used to speed up user interaction by
        stopping the reading of every well sample.
    name : str, optional
    start_time : datetime, optional
        Time when the first image of this acquisition was collected
    well_sample_ref : WellSampleRef, optional
    """

    id: PlateAcquisitionID
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    description: Optional[str] = None
    end_time: Optional[datetime] = None
    maximum_field_count: Optional[PositiveInt] = None
    name: Optional[str] = None
    start_time: Optional[datetime] = None
    well_sample_ref: List[WellSampleRef] = Field(default_factory=list)



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import DichroicID


class DichroicRef(Reference, OMEType):
    """DichroicRef.

    Parameters
    ----------
    id : DichroicID
    """

    id: DichroicID



import io
import re
from xml.etree import ElementTree

# --------------------------------------------------------------------
# Taken from Python 3.8's ElementTree.py.


def canonicalize(xml_data=None, *, out=None, from_file=None, **options):
    """Convert XML to its C14N 2.0 serialised form.

    If *out* is provided, it must be a file or file-like object that receives
    the serialised canonical XML output (text, not bytes) through its ``.write()``
    method.  To write to a file, open it in text mode with encoding "utf-8".
    If *out* is not provided, this function returns the output as text string.

    Either *xml_data* (an XML string) or *from_file* (a file path or
    file-like object) must be provided as input.

    The configuration options are the same as for the ``C14NWriterTarget``.
    """
    if xml_data is None and from_file is None:
        raise ValueError("Either 'xml_data' or 'from_file' must be provided as input")
    sio = None
    if out is None:
        sio = out = io.StringIO()

    parser = ElementTree.XMLParser(target=C14NWriterTarget(out.write, **options))

    if xml_data is not None:
        parser.feed(xml_data)
        parser.close()
    elif from_file is not None:
        ElementTree.parse(from_file, parser=parser)

    return sio.getvalue() if sio is not None else None


_looks_like_prefix_name = re.compile(r"^\w+:\w+$", re.UNICODE).match


class C14NWriterTarget:
    """
    Canonicalization writer target for the XMLParser.

    Serialises parse events to XML C14N 2.0.

    The *write* function is used for writing out the resulting data stream
    as text (not bytes).  To write to a file, open it in text mode with encoding
    "utf-8" and pass its ``.write`` method.

    Configuration options:

    - *with_comments*: set to true to include comments
    - *strip_text*: set to true to strip whitespace before and after text content
    - *rewrite_prefixes*: set to true to replace namespace prefixes by "n{number}"
    - *qname_aware_tags*: a set of qname aware tag names in which prefixes
                          should be replaced in text content
    - *qname_aware_attrs*: a set of qname aware attribute names in which prefixes
                           should be replaced in text content
    - *exclude_attrs*: a set of attribute names that should not be serialised
    - *exclude_tags*: a set of tag names that should not be serialised
    """

    def __init__(
        self,
        write,
        *,
        with_comments=False,
        strip_text=False,
        rewrite_prefixes=False,
        qname_aware_tags=None,
        qname_aware_attrs=None,
        exclude_attrs=None,
        exclude_tags=None,
    ):
        self._write = write
        self._data = []
        self._with_comments = with_comments
        self._strip_text = strip_text
        self._exclude_attrs = set(exclude_attrs) if exclude_attrs else None
        self._exclude_tags = set(exclude_tags) if exclude_tags else None

        self._rewrite_prefixes = rewrite_prefixes
        if qname_aware_tags:
            self._qname_aware_tags = set(qname_aware_tags)
        else:
            self._qname_aware_tags = None
        if qname_aware_attrs:
            self._find_qname_aware_attrs = set(qname_aware_attrs).intersection
        else:
            self._find_qname_aware_attrs = None

        # Stack with globally and newly declared namespaces as (uri, prefix) pairs.
        self._declared_ns_stack = [
            [
                ("http://www.w3.org/XML/1998/namespace", "xml"),
            ]
        ]
        # Stack with user declared namespace prefixes as (uri, prefix) pairs.
        self._ns_stack = []
        if not rewrite_prefixes:
            # this must not be separated from ElementTree
            self._ns_stack.append(list(ElementTree._namespace_map.items()))
        self._ns_stack.append([])
        self._prefix_map = {}
        self._preserve_space = [False]
        self._pending_start = None
        self._root_seen = False
        self._root_done = False
        self._ignored_depth = 0

    def _iter_namespaces(self, ns_stack, _reversed=reversed):
        for namespaces in _reversed(ns_stack):
            if namespaces:  # almost no element declares new namespaces
                yield from namespaces

    def _resolve_prefix_name(self, prefixed_name):
        prefix, name = prefixed_name.split(":", 1)
        for uri, p in self._iter_namespaces(self._ns_stack):
            if p == prefix:
                return f"{{{uri}}}{name}"
        raise ValueError(
            f'Prefix {prefix} of QName "{prefixed_name}" is not declared in scope'
        )

    def _qname(self, qname, uri=None):
        if uri is None:
            uri, tag = qname[1:].rsplit("}", 1) if qname[:1] == "{" else ("", qname)
        else:
            tag = qname

        prefixes_seen = set()
        for u, prefix in self._iter_namespaces(self._declared_ns_stack):
            if u == uri and prefix not in prefixes_seen:
                return f"{prefix}:{tag}" if prefix else tag, tag, uri
            prefixes_seen.add(prefix)

        # Not declared yet => add new declaration.
        if self._rewrite_prefixes:
            if uri in self._prefix_map:
                prefix = self._prefix_map[uri]
            else:
                prefix = self._prefix_map[uri] = f"n{len(self._prefix_map)}"
            self._declared_ns_stack[-1].append((uri, prefix))
            return f"{prefix}:{tag}", tag, uri

        if not uri and "" not in prefixes_seen:
            # No default namespace declared => no prefix needed.
            return tag, tag, uri

        for u, prefix in self._iter_namespaces(self._ns_stack):
            if u == uri:
                self._declared_ns_stack[-1].append((uri, prefix))
                return f"{prefix}:{tag}" if prefix else tag, tag, uri

        raise ValueError(f'Namespace "{uri}" is not declared in scope')

    def data(self, data):
        if not self._ignored_depth:
            self._data.append(data)

    def _flush(self, _join_text="".join):
        data = _join_text(self._data)
        del self._data[:]
        if self._strip_text and not self._preserve_space[-1]:
            data = data.strip()
        if self._pending_start is not None:
            args, self._pending_start = self._pending_start, None
            qname_text = data if data and _looks_like_prefix_name(data) else None
            self._start(*args, qname_text)
            if qname_text is not None:
                return
        if data and self._root_seen:
            self._write(_escape_cdata_c14n(data))

    def start_ns(self, prefix, uri):
        if self._ignored_depth:
            return
        # we may have to resolve qnames in text content
        if self._data:
            self._flush()
        self._ns_stack[-1].append((uri, prefix))

    def start(self, tag, attrs):
        if self._exclude_tags is not None and (
            self._ignored_depth or tag in self._exclude_tags
        ):
            self._ignored_depth += 1
            return
        if self._data:
            self._flush()

        new_namespaces = []
        self._declared_ns_stack.append(new_namespaces)

        if self._qname_aware_tags is not None and tag in self._qname_aware_tags:
            # Need to parse text first to see if it requires a prefix declaration.
            self._pending_start = (tag, attrs, new_namespaces)
            return
        self._start(tag, attrs, new_namespaces)

    def _start(self, tag, attrs, new_namespaces, qname_text=None):
        if self._exclude_attrs is not None and attrs:
            attrs = {k: v for k, v in attrs.items() if k not in self._exclude_attrs}

        qnames = {tag, *attrs}
        resolved_names = {}

        # Resolve prefixes in attribute and tag text.
        if qname_text is not None:
            qname = resolved_names[qname_text] = self._resolve_prefix_name(qname_text)
            qnames.add(qname)
        if self._find_qname_aware_attrs is not None and attrs:
            qattrs = self._find_qname_aware_attrs(attrs)
            if qattrs:
                for attr_name in qattrs:
                    value = attrs[attr_name]
                    if _looks_like_prefix_name(value):
                        qname = resolved_names[value] = self._resolve_prefix_name(value)
                        qnames.add(qname)
            else:
                qattrs = None
        else:
            qattrs = None

        # Assign prefixes in lexicographical order of used URIs.
        parse_qname = self._qname
        parsed_qnames = {
            n: parse_qname(n) for n in sorted(qnames, key=lambda n: n.split("}", 1))
        }

        # Write namespace declarations in prefix order ...
        if new_namespaces:
            attr_list = [
                ("xmlns:" + prefix if prefix else "xmlns", uri)
                for uri, prefix in new_namespaces
            ]
            attr_list.sort()
        else:
            # almost always empty
            attr_list = []

        # ... followed by attributes in URI+name order
        if attrs:
            for k, v in sorted(attrs.items()):
                if qattrs is not None and k in qattrs and v in resolved_names:
                    v = parsed_qnames[resolved_names[v]][0]
                attr_qname, attr_name, uri = parsed_qnames[k]
                # No prefix for attributes in default ('') namespace.
                attr_list.append((attr_qname if uri else attr_name, v))

        # Honour xml:space attributes.
        space_behaviour = attrs.get("{http://www.w3.org/XML/1998/namespace}space")
        self._preserve_space.append(
            space_behaviour == "preserve"
            if space_behaviour
            else self._preserve_space[-1]
        )

        # Write the tag.
        write = self._write
        write("<" + parsed_qnames[tag][0])
        if attr_list:
            write("".join([f' {k}="{_escape_attrib_c14n(v)}"' for k, v in attr_list]))
        write(">")

        # Write the resolved qname text content.
        if qname_text is not None:
            write(_escape_cdata_c14n(parsed_qnames[resolved_names[qname_text]][0]))

        self._root_seen = True
        self._ns_stack.append([])

    def end(self, tag):
        if self._ignored_depth:
            self._ignored_depth -= 1
            return
        if self._data:
            self._flush()
        self._write(f"</{self._qname(tag)[0]}>")
        self._preserve_space.pop()
        self._root_done = len(self._preserve_space) == 1
        self._declared_ns_stack.pop()
        self._ns_stack.pop()

    def comment(self, text):
        if not self._with_comments:
            return
        if self._ignored_depth:
            return
        if self._root_done:
            self._write("\n")
        elif self._root_seen and self._data:
            self._flush()
        self._write(f"<!--{_escape_cdata_c14n(text)}-->")
        if not self._root_seen:
            self._write("\n")

    def pi(self, target, data):
        if self._ignored_depth:
            return
        if self._root_done:
            self._write("\n")
        elif self._root_seen and self._data:
            self._flush()
        self._write(
            f"<?{target} {_escape_cdata_c14n(data)}?>" if data else f"<?{target}?>"
        )
        if not self._root_seen:
            self._write("\n")


def _escape_cdata_c14n(text):
    # escape character data
    try:
        # it's worth avoiding do-nothing calls for strings that are
        # shorter than 500 character, or so.  assume that's, by far,
        # the most common case in most applications.
        if "&" in text:
            text = text.replace("&", "&amp;")
        if "<" in text:
            text = text.replace("<", "&lt;")
        if ">" in text:
            text = text.replace(">", "&gt;")
        if "\r" in text:
            text = text.replace("\r", "&#xD;")
        return text
    except (TypeError, AttributeError):
        ElementTree._raise_serialization_error(text)


def _escape_attrib_c14n(text):
    # escape attribute value
    try:
        if "&" in text:
            text = text.replace("&", "&amp;")
        if "<" in text:
            text = text.replace("<", "&lt;")
        if '"' in text:
            text = text.replace('"', "&quot;")
        if "\t" in text:
            text = text.replace("\t", "&#x9;")
        if "\n" in text:
            text = text.replace("\n", "&#xA;")
        if "\r" in text:
            text = text.replace("\r", "&#xD;")
        return text
    except (TypeError, AttributeError):
        ElementTree._raise_serialization_error(text)



from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .simple_types import Hex40, NonNegativeInt, UnitsLength, UnitsTime


class Plane(OMEType):
    """The Plane object holds microscope stage and image timing data for a given
    channel/z-section/timepoint.

    Parameters
    ----------
    the_c : NonNegativeInt
        The channel this plane is for. This is numbered from 0.
    the_t : NonNegativeInt
        The timepoint this plane is for. This is numbered from 0.
    the_z : NonNegativeInt
        The Z-section this plane is for. This is numbered from 0.
    annotation_ref : AnnotationRef, optional
    delta_t : float, optional
        Time since the beginning of the experiment. Units are set by
        DeltaTUnit.
    delta_t_unit : UnitsTime, optional
        The units of the DeltaT - default:seconds.
    exposure_time : float, optional
        The length of the exposure. Units are set by ExposureTimeUnit.
    exposure_time_unit : UnitsTime, optional
        The units of the ExposureTime - default:seconds.
    hash_sha1 : Hex40, optional
    position_x : float, optional
        The X position of the stage. Units are set by PositionXUnit.
    position_x_unit : UnitsLength, optional
        The units of the X stage position - default:.
    position_y : float, optional
        The Y position of the stage. Units are set by PositionYUnit.
    position_y_unit : UnitsLength, optional
        The units of the Y stage position - default:.
    position_z : float, optional
        The Z position of the stage. Units are set by PositionZUnit.
    position_z_unit : UnitsLength, optional
        The units of the Z stage position - default:.
    """

    the_c: NonNegativeInt
    the_t: NonNegativeInt
    the_z: NonNegativeInt
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    delta_t: Optional[float] = None
    delta_t_unit: Optional[UnitsTime] = UnitsTime("s")
    exposure_time: Optional[float] = None
    exposure_time_unit: Optional[UnitsTime] = UnitsTime("s")
    hash_sha1: Optional[Hex40] = None
    position_x: Optional[float] = None
    position_x_unit: Optional[UnitsLength] = UnitsLength("reference frame")
    position_y: Optional[float] = None
    position_y_unit: Optional[UnitsLength] = UnitsLength("reference frame")
    position_z: Optional[float] = None
    position_z_unit: Optional[UnitsLength] = UnitsLength("reference frame")



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import ProjectID


class ProjectRef(Reference, OMEType):
    """There may be one or more of these in a Dataset.

    This empty element has a required Project ID attribute that refers to Projects
    defined within the OME element.

    Parameters
    ----------
    id : ProjectID
    """

    id: ProjectID



from pathlib import Path

import pytest

nplg = pytest.importorskip("ome_types._napari_plugin")

DATA = Path(__file__).parent / "data"


@pytest.mark.parametrize("fname", DATA.iterdir(), ids=lambda x: x.stem)
def test_widget(fname, qtbot):
    if fname.stem in ("bad.ome", "timestampannotation.ome"):
        pytest.xfail()
    nplg.OMETree(str(fname))



import os
import warnings
from typing import TYPE_CHECKING, Any, Optional, Union

from .model import OME

try:
    from qtpy.QtCore import QMimeData, Qt
    from qtpy.QtWidgets import QTreeWidget, QTreeWidgetItem
except ImportError as e:
    raise ImportError(
        "qtpy and a Qt backend (pyside or pyqt) is required to use the OME widget:\n"
        "pip install qtpy pyqt5"
    ) from e


if TYPE_CHECKING:
    import napari


class OMETree(QTreeWidget):
    """A Widget that can show OME XML."""

    def __init__(
        self, ome_dict: dict = None, viewer: "napari.viewer.Viewer" = None, parent=None
    ) -> None:
        super().__init__(parent=parent)
        self._viewer = viewer
        self.setAcceptDrops(True)
        self.setDropIndicatorShown(True)
        self.setIndentation(15)

        item = self.headerItem()
        font = item.font(0)
        font.setBold(True)
        item.setFont(0, font)
        self.clear()

        self._current_path: Optional[str] = None
        if ome_dict:
            self.update(ome_dict)

        if viewer is not None:
            viewer.layers.selection.events.active.connect(
                lambda e: self._try_load_layer(e.value)
            )
            self._try_load_layer(viewer.layers.selection.active)

    def clear(self) -> None:
        """Clear the widget and reset the header text."""
        self.headerItem().setText(0, "drag/drop file...")
        super().clear()

    def _try_load_layer(self, layer: "napari.layers.Layer"):
        """Handle napari viewer behavior."""
        from ._napari_plugin import METADATA_KEY

        if layer is not None:
            path = str(layer.source.path)

            # deprecated... don't do this ... it should be a dict
            if callable(layer.metadata):
                ome_meta = layer.metadata()
            elif isinstance(layer.metadata, OME):
                ome_meta = layer.metadata
            else:
                ome_meta = layer.metadata.get(METADATA_KEY)
                if callable(ome_meta):
                    ome_meta = ome_meta()

            ome = None
            if isinstance(ome_meta, OME):
                ome = ome_meta
            elif path.endswith((".tiff", ".tif")) and path != self._current_path:
                try:
                    ome = OME.from_tiff(path)
                except Exception:
                    return
            if isinstance(ome, OME):
                self._current_path = path
                self.update(ome)
                self.headerItem().setText(0, os.path.basename(path))
        else:
            self._current_path = None
            self.clear()

    def update(self, ome: Union[OME, str]) -> None:
        """Update the widget with a new OME object or path to an OME XML file."""
        if not ome:
            return
        if isinstance(ome, OME):
            _ome = ome
        elif isinstance(ome, str):
            if ome == self._current_path:
                return
            try:
                if ome.endswith(".xml"):
                    _ome = OME.from_xml(ome)
                elif ome.lower().endswith((".tif", ".tiff")):
                    _ome = OME.from_tiff(ome)
                else:
                    warnings.warn(f"Unrecognized file type: {ome}", stacklevel=2)
                    return
            except Exception as e:
                warnings.warn(
                    f"Could not parse OME metadata from {ome}: {e}", stacklevel=2
                )
                return
            self.headerItem().setText(0, os.path.basename(ome))
            self._current_path = ome
        else:
            raise TypeError("must be OME object or string")
        self._fill_item(_ome.dict(exclude_unset=True))

    def _fill_item(self, obj, item: QTreeWidgetItem = None):
        if item is None:
            self.clear()
            item = self.invisibleRootItem()
        if isinstance(obj, dict):
            for key, val in sorted(obj.items()):
                child = QTreeWidgetItem([key])
                item.addChild(child)
                self._fill_item(val, child)
        elif isinstance(obj, (list, tuple)):
            for n, val in enumerate(obj):
                text = val.get("id", n) if hasattr(val, "get") else n
                child = QTreeWidgetItem([str(text)])
                item.addChild(child)
                self._fill_item(val, child)
        else:
            t = getattr(obj, "value", str(obj))
            item.setText(0, f"{item.text(0)}: {t}")

    def dropMimeData(
        self, parent: QTreeWidgetItem, index: int, data: QMimeData, _: Any
    ) -> bool:
        """Handle drag/drop events to load OME XML files."""
        if data.hasUrls():
            for url in data.urls():
                lf = url.toLocalFile()
                if lf.endswith((".xml", ".tiff", ".tif")):
                    self.update(lf)
                    return True
        return False

    def mimeTypes(self) -> list[str]:
        """Return the supported mime types for drag/drop events."""
        return ["text/uri-list"]

    def supportedDropActions(self) -> "Qt.DropActions":
        """Return the supported drop actions for drag/drop events."""
        return Qt.CopyAction


if __name__ == "__main__":
    from qtpy.QtWidgets import QApplication

    app = QApplication([])

    widget = OMETree()
    widget.show()

    app.exec()



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import ChannelID


class ChannelRef(Reference, OMEType):
    """ChannelRef.

    Parameters
    ----------
    id : ChannelID
    """

    id: ChannelID



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Container, MutableSequence, Union, cast

from typing_extensions import get_args

from . import model
from ._constants import OME_2016_06_XSD, URI_OME
from .model.shape_group import ShapeGroupType
from .util import _ensure_xml_bytes, _get_plural, camel_to_snake, cast_number, norm_key

NEED_INT = [s.__name__ for s in get_args(ShapeGroupType)]
NEED_INT.extend(["Channel", "Well"])


def _is_xml_comment(element: Element) -> bool:
    return False


if TYPE_CHECKING:
    import xml.etree.ElementTree

    import lxml.etree

    Element = Union[xml.etree.ElementTree.Element, lxml.etree._Element]
    ElementTree = Union[xml.etree.ElementTree.ElementTree, lxml.etree._ElementTree]
    Value = Union[float, str, int, bytearray, bool, dict[str, Any], list[Any]]
    Parser = Callable[[bytes], Element]

    XML: Parser
    tostring: Callable[[Element], bytes]
    parse: Callable[[str], ElementTree]

else:
    try:
        # faster if it's available
        from lxml.etree import XML, _Comment, parse, tostring

        def _is_xml_comment(element: Element) -> bool:
            return isinstance(element, _Comment)

    except ImportError:
        from xml.etree.ElementTree import XML, parse, tostring


def elem2dict(node: Element, exclude_null: bool = True) -> dict[str, Any]:
    """Convert an xml.etree or lxml.etree Element into a dict.

    Parameters
    ----------
    node : Element
        The Element to convert. Should be an `xml.etree.ElementTree.Element` or a
        `lxml.etree._Element`
    exclude_null : bool, optional
        If True, exclude keys with null values from the output.


    Returns
    -------
    dict[str, Any]
        The converted Element.
    """
    result: dict[str, Any] = {}

    # Re-used valued
    norm_node = norm_key(node.tag)
    # set of keys that are lists
    norm_list: Container = model._lists.get(norm_node, {})

    # Process attributes
    for key, val in node.attrib.items():
        is_list = key in norm_list
        key = camel_to_snake(norm_key(key))
        if norm_node in NEED_INT:
            val = cast_number(val)
        if is_list:
            key = _get_plural(key, node.tag)
            if key not in result:
                result[key] = []
            cast("list", result[key]).extend(val.split())
        else:
            result[key] = val

    # Process children
    for element in node:
        element = cast("Element", element)
        if _is_xml_comment(element):
            continue
        key = norm_key(element.tag)

        # Process element as tree element if inner XML contains non-whitespace content
        if element.text and element.text.strip():
            value: Any = element.text
            if element.attrib.items():
                value = {"value": value}
                for k, val in element.attrib.items():
                    value[camel_to_snake(norm_key(k))] = val

        elif key == "MetadataOnly":
            value = True

        else:
            value = elem2dict(element, exclude_null=exclude_null)
            if key == "XMLAnnotation":
                value["value"] = tostring(element[0])

        is_list = key in norm_list
        key = camel_to_snake(key)
        if is_list:
            if key == "bin_data" and value["length"] == "0" and "value" not in value:
                value["value"] = ""
            key = _get_plural(key, node.tag)
            if key not in result:
                result[key] = []
            cast("list", result[key]).append(value)

        elif key == "structured_annotations":
            annotations = []
            for _type in (
                "boolean_annotation",
                "comment_annotation",
                "double_annotation",
                "file_annotation",
                "list_annotation",
                "long_annotation",
                "map_annotation",
                "tag_annotation",
                "term_annotation",
                "timestamp_annotation",
                "xml_annotation",
            ):
                if _type in value:
                    values = value.pop(_type)
                    if not isinstance(values, list):
                        values = [values]

                    for v in values:
                        v["_type"] = _type

                        # Normalize empty element to zero-length string.
                        if "value" not in v or v["value"] is None:
                            v["value"] = ""
                    annotations.extend(values)

            if key in result:
                raise ValueError("Duplicate structured_annotations")
            result[key] = annotations

        elif value or not exclude_null:
            try:
                rv = result[key]
            except KeyError:
                result[key] = [value] if key.lower() == "m" else value
            else:
                if not isinstance(rv, MutableSequence) or not rv:
                    result[key] = [rv, value]
                elif isinstance(rv[0], MutableSequence) or not isinstance(
                    value, MutableSequence
                ):
                    rv.append(value)
                else:
                    result[key] = [result, value]

    return result


def validate_lxml(node: Element) -> Element:
    """Ensure that `node` is valid OMX XML.

    Raises
    ------
    lxml.etree.XMLSchemaValidateError
        If `node` is not valid OME XML.
    """
    # TODO: unify with xmlschema validate
    try:
        import lxml.etree
    except ImportError as e:
        raise ImportError("validating xml requires lxml") from e

    for key, val in node.attrib.items():
        if "schemaLocation" in key:
            ns, uri = val.split()
            if ns == URI_OME:
                uri = OME_2016_06_XSD
            schema_doc = parse(uri)  # noqa: S314
            break
    if not lxml.etree.XMLSchema(schema_doc).validate(node):
        raise lxml.etree.XMLSchemaValidateError(
            f"XML did not pass validation error against {uri}"
        )
    return node


def xml2dict(
    path_or_str: Path | str | bytes, validate: bool | None = False
) -> dict[str, Any]:
    """Convert XML string or path to dict using lxml.

    xml : Union[Path, str, bytes]
        XML string, bytes, or path to XML file.
    validate : Optional[bool]
        Whether to validate XML as valid OME XML, by default False.

    Returns
    -------
    Dict[str, Any]
        OME dict
    """
    root = XML(_ensure_xml_bytes(path_or_str))
    if validate:
        validate_lxml(root)

    return elem2dict(root)


def __getattr__(name: str) -> Any:
    """Import lxml if it is not already imported."""
    if name == "lxml2dict":
        import warnings

        warnings.warn(
            "lxml2dict is deprecated, use xml2dict instead",
            FutureWarning,
            stacklevel=2,
        )
        return xml2dict
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



import pint

ureg = pint.UnitRegistry(auto_reduce_dimensions=True)
ureg.define("reference_frame = [_reference_frame]")
ureg.define("@alias grade = gradian")
ureg.define("@alias astronomical_unit = ua")
ureg.define("line = inch / 12")
ureg.define("millitorr = torr / 1000 = mTorr")
ureg.define("@alias torr = Torr")



from ome_types._base_type import OMEType

from .reference import Reference
from .simple_types import FilterSetID


class FilterSetRef(Reference, OMEType):
    """FilterSetRef.

    Parameters
    ----------
    id : FilterSetID
    """

    id: FilterSetID



from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .annotation_ref import AnnotationRef
from .manufacturer_spec import ManufacturerSpec
from .simple_types import LightSourceID, UnitsPower


class LightSource(ManufacturerSpec, OMEType):
    """The lightsource for the instrument.

    An instrument may have several light sources. The type of lightsource is
    specified by one of the child-elements which are 'Laser', 'Filament', 'Arc' or
    'LightEmittingDiode'. Each of the light source types has its own Type
    attribute to further differentiate the light source (eg, Nd-YAG for Laser or
    Hg for Arc).

    Parameters
    ----------
    id : LightSourceID
        A LightSource ID must be specified for each light source, and the
        individual light sources can be referred to by their LightSource IDs
        (eg from Channel).
    annotation_ref : AnnotationRef, optional
    lot_number : str, optional
        The lot number of the component.
    manufacturer : str, optional
        The manufacturer of the component.
    model : str, optional
        The Model of the component.
    power : float, optional
        The light-source power. Units are set by PowerUnit.
    power_unit : UnitsPower, optional
        The units of the Power - default:milliwatts.
    serial_number : str, optional
        The serial number of the component.
    """

    id: LightSourceID
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    power: Optional[float] = None
    power_unit: Optional[UnitsPower] = UnitsPower("mW")



from enum import Enum
from typing import List, Optional

from pydantic import Field

from ome_types._base_type import OMEType

from .affine_transform import AffineTransform
from .annotation_ref import AnnotationRef
from .simple_types import Color, NonNegativeInt, ShapeID, UnitsLength


class FillRule(Enum):
    EVEN_ODD = "EvenOdd"
    NON_ZERO = "NonZero"


class FontFamily(Enum):
    CURSIVE = "cursive"
    FANTASY = "fantasy"
    MONOSPACE = "monospace"
    SANSSERIF = "sans-serif"
    SERIF = "serif"


class FontStyle(Enum):
    BOLD = "Bold"
    BOLD_ITALIC = "BoldItalic"
    ITALIC = "Italic"
    NORMAL = "Normal"


class Shape(OMEType):
    """The shape element contains a single specific ROI shape and links that to any
    channels, and a timepoint and a z-section.

    It also records any transform applied to the ROI shape.

    Parameters
    ----------
    id : ShapeID
    annotation_ref : AnnotationRef, optional
    fill_color : Color, optional
        The color of the fill - encoded as RGBA The value "-1" is #FFFFFFFF so
        solid white (it is a signed 32 bit value) NOTE: Prior to the 2012-06
        schema the default value was incorrect and produced a transparent red
        not solid white.
    fill_rule : FillRule, optional
    font_family : FontFamily, optional
    font_size : NonNegativeInt, optional
        Size of the font. Units are set by FontSizeUnit.
    font_size_unit : UnitsLength, optional
        The units used for the font size.
    font_style : FontStyle, optional
    locked : bool, optional
        Controls whether the shape is locked and read only, true is locked,
        false is editable.
    stroke_color : Color, optional
        The color of the stroke  - encoded as RGBA The value "-1" is #FFFFFFFF
        so solid white (it is a signed 32 bit value) NOTE: Prior to the
        2012-06 schema the default value was incorrect and produced a
        transparent red not solid white.
    stroke_dash_array : str, optional
        e.g. "none", "10 20 30 10"
    stroke_width : float, optional
        The width of the stroke. Units are set by StrokeWidthUnit.
    stroke_width_unit : UnitsLength, optional
        The units used for the stroke width.
    text : str, optional
    the_c : NonNegativeInt, optional
        The channel the ROI applies to. If not specified then the ROI applies
        to all the channels of the image. This is numbered from 0.
    the_t : NonNegativeInt, optional
        The timepoint the ROI applies to. If not specified then the ROI
        applies to all the timepoints of the image. This is numbered from 0.
    the_z : NonNegativeInt, optional
        The z-section the ROI applies to. If not specified then the ROI
        applies to all the z-sections of the image. This is numbered from 0.
    transform : AffineTransform, optional
        This is a matrix used to transform the shape. The element has 6
        xsd:float attributes. If the element is present then all 6 values must
        be included.
    """

    id: ShapeID
    annotation_ref: List[AnnotationRef] = Field(default_factory=list)
    fill_color: Optional[Color] = None
    fill_rule: Optional[FillRule] = None
    font_family: Optional[FontFamily] = None
    font_size: Optional[NonNegativeInt] = None
    font_size_unit: Optional[UnitsLength] = UnitsLength("pt")
    font_style: Optional[FontStyle] = None
    locked: Optional[bool] = None
    stroke_color: Optional[Color] = None
    stroke_dash_array: Optional[str] = None
    stroke_width: Optional[float] = None
    stroke_width_unit: Optional[UnitsLength] = UnitsLength("pixel")
    text: Optional[str] = None
    the_c: Optional[NonNegativeInt] = None
    the_t: Optional[NonNegativeInt] = None
    the_z: Optional[NonNegativeInt] = None
    transform: Optional[AffineTransform] = None



from enum import Enum
from typing import Optional

from ome_types._base_type import OMEType

from .settings import Settings
from .simple_types import ObjectiveID


class Medium(Enum):
    AIR = "Air"
    GLYCEROL = "Glycerol"
    OIL = "Oil"
    OTHER = "Other"
    WATER = "Water"


class ObjectiveSettings(Settings, OMEType):
    """This holds the setting applied to an objective as well as a reference to the
    objective.

    The ID is the objective used in this case.

    Parameters
    ----------
    id : ObjectiveID
    correction_collar : float, optional
        The CorrectionCollar is normally an adjustable ring on the objective.
        Each has an arbitrary scale on it so the values is unit-less.
    medium : Medium, optional
    refractive_index : float, optional
        The RefractiveIndex is that of the immersion medium. This is a ratio
        so it also unit-less.
    """

    id: ObjectiveID
    correction_collar: Optional[float] = None
    medium: Optional[Medium] = None
    refractive_index: Optional[float] = None



from typing import Optional, cast

from ome_types._base_type import OMEType

from .simple_types import NonNegativeInt, UniversallyUniqueIdentifier


class UUID(OMEType):
    file_name: str
    value: UniversallyUniqueIdentifier


class TiffData(OMEType):
    """This described the location of the pixel data in a tiff file.

    Parameters
    ----------
    first_c : NonNegativeInt, optional
        Gives the C position of the image plane at the specified IFD. Indexed
        from 0. Default is 0 (the first C position).
    first_t : NonNegativeInt, optional
        Gives the T position of the image plane at the specified IFD. Indexed
        from 0. Default is 0 (the first T position).
    first_z : NonNegativeInt, optional
        Gives the Z position of the image plane at the specified IFD. Indexed
        from 0. Default is 0 (the first Z position).
    ifd : NonNegativeInt, optional
        Gives the IFD(s) for which this element is applicable. Indexed from 0.
        Default is 0 (the first IFD).
    plane_count : NonNegativeInt, optional
        Gives the number of IFDs affected. Dimension order of IFDs is given by
        the enclosing Pixels element's DimensionOrder attribute. Default is
        the number of IFDs in the TIFF file, unless an IFD is specified, in
        which case the default is 1.
    uuid : Optional[UUID], optional
        This must be used when the IFDs are located in another file. Note: It
        is permissible for this to be self referential.
    """

    first_c: Optional[NonNegativeInt] = cast(NonNegativeInt, 0)
    first_t: Optional[NonNegativeInt] = cast(NonNegativeInt, 0)
    first_z: Optional[NonNegativeInt] = cast(NonNegativeInt, 0)
    ifd: Optional[NonNegativeInt] = cast(NonNegativeInt, 0)
    plane_count: Optional[NonNegativeInt] = None
    uuid: Optional[UUID] = None



from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, Sequence, Set

from pydantic import BaseModel, validator

if TYPE_CHECKING:
    import pint


class Sentinel:
    """Create singleton sentinel objects with a readable repr."""

    def __init__(self, name: str) -> None:
        self.name = name

    def __repr__(self) -> str:
        return f"{__name__}.{self.name}.{id(self)}"


def quantity_property(field: str) -> property:
    """Create property that returns a ``pint.Quantity`` combining value and unit."""

    def quantity(self: Any) -> Optional["pint.Quantity"]:
        from ome_types._units import ureg

        value = getattr(self, field)
        if value is None:
            return None
        unit = getattr(self, f"{field}_unit").value.replace(" ", "_")
        return ureg.Quantity(value, unit)

    return property(quantity)


class OMEType(BaseModel):
    """The base class that all OME Types inherit from.

    This provides some global conveniences around auto-setting ids. (i.e., making them
    optional in the class constructor, but never ``None`` after initialization.).
    It provides a nice __repr__ that hides things that haven't been changed from
    defaults.  It adds ``*_quantity`` property for fields that have both a value and a
    unit, where ``*_quantity`` is a pint ``Quantity``.  It also provides pickling
    support.
    """

    # Default value to support automatic numbering for id field values.
    _AUTO_SEQUENCE = Sentinel("AUTO_SEQUENCE")
    # allow use with weakref
    __slots__: ClassVar[Set[str]] = {"__weakref__"}  # type: ignore

    def __init__(__pydantic_self__, **data: Any) -> None:
        if "id" in __pydantic_self__.__fields__:
            data.setdefault("id", OMEType._AUTO_SEQUENCE)
        super().__init__(**data)

    def __init_subclass__(cls) -> None:
        """Add some properties to subclasses with units.

        It adds ``*_quantity`` property for fields that have both a value and a
        unit, where ``*_quantity`` is a pint ``Quantity``
        """
        _clsdir = set(cls.__fields__)
        for field in _clsdir:
            if f"{field}_unit" in _clsdir:
                setattr(cls, f"{field}_quantity", quantity_property(field))

    # pydantic BaseModel configuration.
    # see: https://pydantic-docs.helpmanual.io/usage/model_config/
    class Config:
        # whether to allow arbitrary user types for fields (they are validated
        # simply by checking if the value is an instance of the type). If
        # False, RuntimeError will be raised on model declaration
        arbitrary_types_allowed = False
        # whether to perform validation on assignment to attributes
        validate_assignment = True
        # whether to treat any underscore non-class var attrs as private
        # https://pydantic-docs.helpmanual.io/usage/models/#private-model-attributes
        underscore_attrs_are_private = True
        # whether to populate models with the value property of enums, rather
        # than the raw enum. This may be useful if you want to serialise
        # model.dict() later.  False by default
        # see conversation in https://github.com/tlambert03/ome-types/pull/74
        use_enum_values = False
        # whether to validate field defaults (default: False)
        validate_all = True

    def __repr__(self: Any) -> str:
        from datetime import datetime
        from enum import Enum
        from textwrap import indent

        name = self.__class__.__qualname__
        lines = []
        for f in sorted(
            self.__fields__.values(), key=lambda f: f.name not in ("name", "id")
        ):
            if f.name.endswith("_"):
                continue
            # https://github.com/python/mypy/issues/6910
            if f.default_factory:
                default = f.default_factory()
            else:
                default = f.default

            current = getattr(self, f.name)
            if current != default:
                if isinstance(current, Sequence) and not isinstance(current, str):
                    rep = f"[<{len(current)} {f.name.title()}>]"
                elif isinstance(current, Enum):
                    rep = repr(current.value)
                elif isinstance(current, datetime):
                    rep = f"datetime.fromisoformat({current.isoformat()!r})"
                else:
                    rep = repr(current)
                lines.append(f"{f.name}={rep},")
        if len(lines) == 1:
            body = lines[-1].rstrip(",")
        elif lines:
            body = "\n" + indent("\n".join(lines), "   ") + "\n"
        else:
            body = ""
        out = f"{name}({body})"
        return out

    @validator("id", pre=True, always=True, check_fields=False)
    def validate_id(cls, value: Any) -> str:
        """Pydantic validator for ID fields in OME models.

        If no value is provided, this validator provides and integer ID, and stores the
        maximum previously-seen value on the class.
        """
        from typing import ClassVar

        # get the required LSID field from the annotation
        id_field = cls.__fields__.get("id")
        if not id_field:
            return value

        # Store the highest seen value on the class._max_id attribute.
        if not hasattr(cls, "_max_id"):
            cls._max_id = 0  # type: ignore [misc]
            cls.__annotations__["_max_id"] = ClassVar[int]
        if value is OMEType._AUTO_SEQUENCE:
            value = cls._max_id + 1
        if isinstance(value, int):
            v_id = value
            id_string = id_field.type_.__name__[:-2]
            value = f"{id_string}:{value}"
        else:
            value = str(value)
            v_id = value.rsplit(":", 1)[-1]
        try:
            v_id = int(v_id)
            cls._max_id = max(cls._max_id, v_id)  # type: ignore [misc]
        except ValueError:
            pass

        return id_field.type_(value)

    def __getstate__(self: Any) -> Dict[str, Any]:
        """Support pickle of our weakref references."""
        state = super().__getstate__()
        state["__private_attribute_values__"].pop("_ref", None)
        return state

    @classmethod
    def snake_name(cls) -> str:
        from .model import _camel_to_snake

        return _camel_to_snake[cls.__name__]
